import numpy as np
import pandas as pd
from scipy.ndimage import binary_fill_holes

from baby.io import load_tiled_image
from baby.tracker.core import CellTracker


class CellBenchmarker:  # TODO Simplify this by inheritance
    """
    Takes a metadata dataframe and a model and estimates the prediction in a trap-wise manner.

    This class can also produce confusion matrices for a given Tracker and validation dataset.
    """

    def __init__(self, meta, model, bak_model, nstepsback=None):
        self.indices = ["experimentID", "position", "trap", "tp"]
        self.cindices = self.indices + ["cellLabels"]
        self.meta = meta.copy()
        self.meta["cont_list_index"] = (*range(len(self.meta)),)

        self.tracker = CellTracker(model=model, bak_model=bak_model)
        if nstepsback is None:
            self.nstepsback = self.tracker.nstepsback
        self.traps_loc

    @property
    def traps_loc(self):
        """
        Generates a list of trap locations using the metadata.
        """
        if not hasattr(self, "_traps_loc"):

            traps = np.unique(
                [
                    ind[: self.indices.index("trap") + 1]
                    for ind in self.meta.index
                ],
                axis=0,
            )
            # str->int conversion
            traps = [(ind[0], *map(int, ind[1:])) for ind in traps]
            self._traps_loc = (*map(tuple, traps),)

        return self._traps_loc

    @property
    def masks(self):
        if not hasattr(self, "_masks"):
            self._masks = [
                load_tiled_image(fname)[0] for fname in self.meta["filename"]
            ]
            for i, mask in enumerate(self._masks):
                for j in range(mask.shape[2]):
                    self._masks[i][..., j] = binary_fill_holes(
                        self._masks[i][..., j]
                    )

            self._masks = [np.moveaxis(mask, 2, 0) for mask in self._masks]

        return self._masks

    def predict_lbls_from_tpimgs(self, tp_img_tuple):
        max_lbl = 0
        prev_feats = []
        cell_lbls = []
        for tp, masks in tp_img_tuple:
            lastn_lbls = cell_lbls[-self.nstepsback :]
            lastn_feats = prev_feats[-self.nstepsback :]
            new_lbl, feats, max_lbl = self.tracker.get_new_lbls(
                masks, lastn_lbls, lastn_feats, max_lbl
            )
            cell_lbls = cell_lbls + [new_lbl]
            prev_feats = prev_feats + [feats]

        return (tp, cell_lbls)

    def df_get_imglist(self, exp, pos, trap, tp=None):
        df = self.meta.loc[(exp, pos, trap), ["cont_list_index", "cellLabels"]]
        return zip(df.index, [self.masks[i] for i in df["cont_list_index"]])

    def predict_set(self, exp, pos, trap, tp=None):
        """
        Predict labels using tp1-tp2 accuracy of prediction
        """
        # print("Processing trap {}".format(exp, pos, trap))
        tp_img_tuple = (*self.df_get_imglist(exp, pos, trap),)
        tp, lbl_list = self.predict_lbls_from_tpimgs(tp_img_tuple)
        # print("loc {}, {}, {}, labels: {}".format(exp, pos, trap, lbl_list))
        return lbl_list

    def compare_traps(self, exp, pos, trap):
        """
        Error calculator for testing model and assignment heuristics.

        Uses the trap id to compare the amount of cells correctly predicted.
        This uses local indices, not whole timepoints. It returns the
        fraction of cells correctly predicted, and the timepoints of mistakes

        Returns:
        float: Fraction of cells correctly predicted
        list of 2-sized tuples: list of tp id of errors and the mistaken cell

        """
        print("Processing trap {}, {}, {}".format(exp, pos, trap))
        new_cids = self.predict_set(exp, pos, trap)

        test_df = self.meta.loc(axis=0)[(exp, pos, trap)].copy()
        test_df["pred_cellLabels"] = new_cids

        orig = test_df["cellLabels"].values
        new = test_df["pred_cellLabels"].values
        local_indices = [[], []]

        # Case just defines if it is the test or new set
        # print("Making tp-wise comparison")
        for i, case in enumerate(
            (zip(orig[:-1], orig[1:]), zip(new[:-1], new[1:]))
        ):
            for prev_cells, pos_cells in case:
                local_assignment = [
                    prev_cells.index(cell) if cell in prev_cells else -1
                    for cell in pos_cells
                ]
                local_indices[i] += local_assignment

        # Flatten
        if len(local_indices) > 2:
            flt_test, flt_new = [
                np.array([j for i in case for j in i])
                for case in local_indices
            ]
            tp_list = np.array(
                [i for i, vals in enumerate(local_indices[0]) for j in vals]
            )
        else:
            flt_test, flt_new = [
                np.array([i for i in case]) for case in local_indices
            ]
            # tp_list = np.array(
            #     [i for i, vals in enumerate(local_indices[0]) for j in vals])

        correct = flt_test == flt_new
        if len(local_indices) > 2:
            error_list = tp_list[~correct]
        error_cid = (
            test_df.iloc[1:]["cellLabels"].explode().dropna()[~correct].values
        )
        frac_correct = np.mean(correct)

        print("Fraction of correct predictions", frac_correct)
        if len(local_indices) > 2:
            return (frac_correct, list(zip(error_list, error_cid)))
        else:
            # print("Warning: Single set of tps for this position")
            return (frac_correct, error_cid)

    def predict_all(self):
        """
        Predict all datasets defined in self.traps_loc
        """
        stepsback = [2]
        threshs = [0.9]
        self.predictions = {}
        for nstepsback in stepsback:
            for thresh in threshs:
                self.nstepsback = nstepsback
                self.tracker.nstepsback = nstepsback
                self.low_thresh = 1 - thresh
                self.high_thresh = thresh
                self.thresh = thresh * 5 / 8
                for address in self.traps_loc:
                    self.predictions[
                        (nstepsback, thresh, address)
                    ] = self.predict_set(*address)

    def calculate_errsum(self):
        """
        Calculate all errors, addresses of images with errors and error fractions.
        """
        frac_errs = {}
        all_errs = {}
        nerrs = {}
        stepsback = list(range(1, 3))
        threshs = [0.95]
        for nstepsback in stepsback:
            for thresh in threshs:
                self.nstepsback = nstepsback
                self.tracker.nstepsback = nstepsback
                self.low_thresh = 1 - thresh
                self.high_thresh = thresh
                self.thresh = thresh * 5 / 8
                all_errs[(thresh, nstepsback)] = {}
                frac_errs[(thresh, nstepsback)] = []
                nerrs[(thresh, nstepsback)] = []
                for address in self.traps_loc:
                    fraction, errors = self.compare_traps(*address)
                    if len(errors):
                        all_errs[(thresh, nstepsback)][address] = errors
                        frac_errs[(thresh, nstepsback)].append(fraction)
                        nerrs[(thresh, nstepsback)].append(len(errors))
                    else:
                        nerrs[(thresh, nstepsback)].append(0)
                        frac_errs[(thresh, nstepsback)].append(1.0)

        return (frac_errs, all_errs, nerrs)

    def get_truth_matrix_from_pair(self, pair):
        """
         Requires self.meta

         args:
         :pair: tuple of size 4 (experimentID, position, trap (tp1, tp2))

         returns

        :truth_mat: boolean ndarray of shape (ncells(tp1) x ncells(tp2)
         links cells in tp1 to cells in tp2
        """

        clabs1 = self.meta.loc[pair[:3] + (pair[3][0],), "cellLabels"]
        clabs2 = self.meta.loc[pair[:3] + (pair[3][1],), "cellLabels"]

        truth_mat = gen_boolmat_from_clabs(clabs1, clabs2)

        return truth_mat

    def get_mota_stats(self, pair):
        true_mat = self.get_truth_matrix_from_pair(pair)
        prob_mat = self.tracker.predict_proba_from_ndarray(
            ndarray, *args, **kwargs
        )
        pred_mat = prob_mat > thresh

        true_flat = true_mat.flatten()
        pred_flat = pred_mat.flatten()

        true_pos = np.sum(true_flat & pred_flat)
        false_pos = np.sum(true_flat & ~pred_flat)

        # TODO add identity switch

    def gen_cm_stats(self, pair, thresh=0.7, *args, **kwargs):
        """
        Calculate confusion matrix for a pair of pos-timepoints
        """

        masks = [self.masks[i] for i in self.meta.loc[pair, "cont_list_index"]]
        feats = [self.tracker.calc_feats_from_mask(mask) for mask in masks]
        ndarray = self.tracker.calc_feat_ndarray(*feats)
        self.tracker.low_thresh = 1 - thresh
        self.tracker.high_thresh = thresh
        prob_mat = self.tracker.predict_proba_from_ndarray(
            ndarray, *args, **kwargs
        )
        pred_mat = prob_mat > thresh

        true_mat = self.get_truth_matrix_from_pair(pair)
        if not len(true_mat) and not len(pred_mat):
            return (0, 0, 0, 0)

        true_flat = true_mat.flatten()
        pred_flat = pred_mat.flatten()

        true_pos = np.sum(true_flat & pred_flat)
        false_pos = np.sum(true_flat & ~pred_flat)
        false_neg = np.sum(~true_flat & pred_flat)
        true_neg = np.sum(~true_flat & ~pred_flat)

        return (true_pos, false_pos, false_neg, true_neg)

    def extract_pairs_from_trap(self, trap_loc):
        subdf = self.meta[["list_index", "cellLabels"]].loc(axis=0)[trap_loc]
        pairs = [
            trap_loc + tuple((pair,))
            for pair in zip(subdf.index[:-1], subdf.index[1:])
        ]

        return pairs

    def gen_pairlist(self):
        self.pairs = [
            self.extract_pairs_from_trap(trap) for trap in self.traps_loc
        ]

    def gen_cm_from_pairs(self, thresh=0.5, *args, **kwargs):
        con_mat = {}
        con_mat["tp"] = 0
        con_mat["fp"] = 0
        con_mat["fn"] = 0
        con_mat["tn"] = 0
        for pairset in self.pairs:
            for pair in pairset:
                res = self.gen_cm_stats(pair, thresh=thresh, *args, **kwargs)
                con_mat["tp"] += res[0]
                con_mat["fp"] += res[1]
                con_mat["fn"] += res[2]
                con_mat["tn"] += res[3]
        self._con_mat = con_mat
        return self._con_mat

    def get_frac_error_df(self):
        """
        Calculates the trap-wise error and averages across a position.
        """
        self.frac_errs, self.all_errs, self.nerrs = self.calculate_errsum()

        # nerrs_df = pd.DataFrame(self.nerrs).melt()
        frac_df = pd.DataFrame(self.frac_errs).melt()

        return frac_df

    def gen_errorplots(self):
        """
        Calculates the trap-wise error and averages across a position.
        """
        frac_df = self.get_frac_error_df()
        import seaborn as sns
        from matplotlib import pyplot as plt

        # ax = sns.barplot(x='variable_0', y='value', data=frac_df)
        ax = sns.barplot(
            x="variable_1", y="value", hue="variable_0", data=frac_df
        )
        ax.set(
            xlabel="Backtrace depth",
            ylabel="Fraction of correct assignments",
            ylim=(0.9, 1),
        )
        plt.legend(title="Threshold")
        plt.savefig("tracker_benchmark_btdepth.png")
        plt.show()

    # def plot_pair(self, address)


def gen_boolmat_from_clabs(clabs1, clabs2):
    if not np.any(clabs1) and not np.any(clabs2):
        return np.array([])

    boolmat = np.zeros((len(clabs1), len(clabs2))).astype(bool)
    for i, lab1 in enumerate(clabs1):
        for j, lab2 in enumerate(clabs2):
            if lab1 == lab2:
                boolmat[i, j] = True

    return boolmat


def gen_stats_dict(results):
    """
    Generates a dictionary using results from different binary classification tasks,
    for example, using different thresholds

    output

    dictionary containing the name of statistic as a key and a list
    of that statistic for the data subsets.
    """
    funs = (get_precision, get_recall, get_tnr, get_balanced_acc)
    names = ("precision", "recall", "TNR", "balanced_acc")
    stats_dict = {
        name: [fun(res) for res in results] for fun, name in zip(funs, names)
    }

    return stats_dict


def get_precision(res_dict):
    return (res_dict["tp"]) / (res_dict["tp"] + res_dict["fp"])


def get_recall(res_dict):
    return res_dict["tp"] / (res_dict["tp"] + res_dict["fn"])


def get_tnr(res_dict):
    return res_dict["tn"] / (res_dict["tn"] + res_dict["fp"])


def get_balanced_acc(res_dict):
    return (get_recall(res_dict) + get_tnr(res_dict)) / 2



# TODO Module docstring
import typing as t
from abc import abstractmethod

import numpy as np
import pandas as pd

from agora.abc import ParametersABC
from agora.utils.kymograph import get_index_as_np
from postprocessor.core.abc import PostProcessABC


class LineageProcessParameters(ParametersABC):
    """
    Parameters
    """

    _defaults = {}


class LineageProcess(PostProcessABC):
    """
    Lineage process that must be passed a (N,3) lineage matrix (where the coliumns are trap, mother, daughter respectively)
    """

    def __init__(self, parameters: LineageProcessParameters):
        super().__init__(parameters)

    @abstractmethod
    def run(
        self,
        signal: pd.DataFrame,
        lineage: np.ndarray,
        *args,
    ):
        pass

    @classmethod
    def as_function(
        cls,
        data: pd.DataFrame,
        lineage: t.Union[t.Dict[t.Tuple[int], t.List[int]]],
        *extra_data,
        **kwargs,
    ):
        """
        Overrides PostProcess.as_function classmethod.
        Lineage functions require lineage information to be passed if run as function.
        """
        parameters = cls.default_parameters(**kwargs)
        return cls(parameters=parameters).run(
            data, lineage=lineage, *extra_data
        )

    def get_lineage_information(self, signal):
        if "mother_label" in signal.index.names:
            lineage = get_index_as_np(signal)
        elif self.cells is not None:
            lineage = self.cells.mothers_daughters
        else:
            raise Exception("No linage information found")
        return lineage



"""
Utility functions and classes
"""
import itertools
import logging
import operator
import typing as t
from functools import wraps
from pathlib import Path
from time import perf_counter
from typing import Callable

import cv2


def repr_obj(obj, indent=0):
    """
    Helper function to display info about OMERO objects.
    Not all objects will have a "name" or owner field.
    """
    string = """%s%s:%s  Name:"%s" (owner=%s)""" % (
        " " * indent,
        obj.OMERO_CLASS,
        obj.getId(),
        obj.getName(),
        obj.getAnnotation(),
    )

    return string


def imread(path):
    return cv2.imread(str(path), -1)


class Cache:
    """
    Fixed-length mapping to use as a cache.
    Deletes items in FIFO manner when maximum allowed length is reached.
    """

    def __init__(self, max_len=5000, load_fn: Callable = imread):
        """
        :param max_len: Maximum number of items in the cache.
        :param load_fn: The function used to load new items if they are not
        available in the Cache
        """
        self._dict = dict()
        self._queue = []
        self.load_fn = load_fn
        self.max_len = max_len

    def __getitem__(self, item):
        if item not in self._dict:
            self.load_item(item)
        return self._dict[item]

    def load_item(self, item):
        self._dict[item] = self.load_fn(item)
        # Clean up the queue
        self._queue.append(item)
        if len(self._queue) > self.max_len:
            del self._dict[self._queue.pop(0)]

    def clear(self):
        self._dict.clear()
        self._queue.clear()


def accumulate(list_: list) -> t.Generator:
    """Accumulate list based on the first value"""
    list_ = sorted(list_)
    it = itertools.groupby(list_, operator.itemgetter(0))
    for key, sub_iter in it:
        yield key, [x[1] for x in sub_iter]


def get_store_path(save_dir, store, name):
    """Create a path to a position-specific store.

    This combines the name and the store's base name into a file path within save_dir.
    For example.
    >>> get_store_path('data', 'baby_seg.h5', 'pos001')
    Path(data/pos001baby_seg.h5')

    :param save_dir: The root directory in which to save the file, absolute
    path.
    :param store: The base name of the store
    :param name: The name of the position
    :return: Path(save_dir) / name+store
    """
    store = Path(save_dir) / store
    store = store.with_name(name + store.name)
    return store


def parametrized(dec):
    def layer(*args, **kwargs):
        def repl(f):
            return dec(f, *args, **kwargs)

        return repl

    return layer


@parametrized
def timed(f, name=None):
    @wraps(f)
    def decorated(*args, **kwargs):
        t = perf_counter()
        res = f(*args, **kwargs)
        to_print = name or f.__name__
        logging.debug(f"Timing:{to_print}:{perf_counter() - t}s")
        return res

    return decorated



#!/usr/bin/env jupyter

import pandas as pd
from agora.abc import ParametersABC

from postprocessor.core.abc import PostProcessABC


class interpolateParameters(ParametersABC):
    """
    Parameters
    """

    _defaults = {"limit_area": "inside"}


class interpolate(PostProcessABC):
    """
    Interpolate process.
    """

    def __init__(self, parameters: interpolateParameters):
        super().__init__(parameters)

    def run(self, signal: pd.DataFrame):
        if len(signal):
            signal = signal.interpolate(limit_area="inside", axis=1)
        return signal



from agora.abc import ParametersABC

from postprocessor.core.abc import PostProcessABC
from postprocessor.core.functions.tracks import get_joinable


class MergerParameters(ParametersABC):
    """
    :param tol: float or int threshold of average (prediction error/std) necessary
        to consider two tracks the same. If float is fraction of first track,
        if int it is absolute units.
    :param window: int value of window used for savgol_filter
    :param degree: int value of polynomial degree passed to savgol_filter
    """

    _defaults = {
        "smooth": False,
        "tolerance": 0.2,
        "window": 5,
        "degree": 3,
        "min_avg_delta": 0.5,
    }


class Merger(PostProcessABC):
    """
    Combines rows of tracklet that are likely to be the same.
    """

    def __init__(self, parameters):
        super().__init__(parameters)

    def run(self, signal):
        joinable = []
        if signal.shape[1] > 4:
            joinable = get_joinable(signal, tol=self.parameters.tolerance)
        return joinable



# -*- coding: utf-8 -*-

import json
import pkgutil
import re
import typing as t
from datetime import datetime
from os.path import dirname, exists, join

CONFIG_KEY = "@@CONFIG@@"
DEFAULT_NOSKIP = {"regex", "regexs", "list", "lists"}
DEFAULT_NOT_USE_UNMATCHED = {"regex", "regexs"}


class GrammarNotFound(OSError):
    pass


class ParseError(Exception):
    pass


class Parser(object):
    def __init__(self, grammar_filename):
        """Create a Parser object based on the grammar defined in a file

        :param grammar_filename: path to json file specifying grammar for this
        parser, or one of the default grammars included with the package
        """

        if exists(grammar_filename):
            with open(grammar_filename, "r") as f:
                self.grammar = json.load(f)
        else:
            if not grammar_filename.endswith(".json"):
                grammar_filename = grammar_filename + ".json"
            try:
                grammar_fd = pkgutil.get_data(
                    __package__, "grammars/" + grammar_filename
                )
            except FileNotFoundError as e:
                raise GrammarNotFound(
                    "{}:specified grammar could not be found:".format(e)
                )
            self.grammar = json.loads(grammar_fd)

        self._config = self.grammar.get(CONFIG_KEY, {})
        if CONFIG_KEY in self.grammar:
            del self.grammar[CONFIG_KEY]

        # Preprocessing to be applied to each line before checking triggers
        self._preprocessing = self._config.get("regex_preprocessing", [])
        self._preprocessing = [re.compile(r) for r in self._preprocessing]

        self._triggers = {
            trigger_type: [
                (k, v[f"trigger_{trigger_type}"])
                if trigger_type != "re"
                else (k, re.compile(v[f"trigger_{trigger_type}"]))
                for k, v in self.grammar.items()
                if f"trigger_{trigger_type}" in v
            ]
            for trigger_type in ("startswith", "endswith", "contains", "re")
        }

    def _set_section(self, k=None):
        if k in self.grammar:
            self._active_section = self.grammar[k]
            self._section_name = k
            self._section_type = self._active_section.get("type")
        else:
            self._active_section = None
            self._section_name = ""
            self._section_type = None

    def parse(self, filehandle):
        """Parse contents of file according to the loaded grammar

        :param filehandle: a line generator, e.g., a valid file handle
        """

        self._set_section()
        table_header = []
        column_types = []

        output = {}

        for line in filehandle:
            line = line.strip()
            if len(line) == 0:
                # skip blank lines
                continue

            line_pp = [r.findall(line) for r in self._preprocessing]
            line_pp = [m[0].strip() for m in line_pp if len(m) == 1]
            line_unmatched = line_pp[0] if len(line_pp) == 1 else line
            line_pp += [line]

            trigger_check_methods = {
                k: lam
                for k, lam in zip(
                    self._triggers.keys(),
                    (
                        lambda x, t: x.startswith(t),
                        lambda x, t: x.endswith(t),
                        lambda x, t: x.find(t),
                        lambda x, re: re.findall(x),
                    ),
                )
            }
            matches = {
                trigger: [
                    (k, trig_str)
                    for k, trig_str in self._triggers[trigger]
                    if any(
                        [
                            trigger_check_methods[trigger](line, trig_str)
                            for line in line_pp
                        ]
                    )
                ]
                for trigger, method in trigger_check_methods.items()
            }
            section_match = {
                k
                for trigger_matches in matches.values()
                for k, _ in trigger_matches
            }

            # if len(section_match) > 1:
            assert len(section_match) <= 1, ParseError(
                "conflicting sections triggered"
            )

            if len(section_match) == 1:
                # Update the active section
                self._set_section(list(section_match)[0])

                # Determine the unmatched part of the line
                line_unmatched = self.determine_unmatched_part(
                    matches, line_pp
                )

                # Skip the matched line if requested
                if self._active_section.get(
                    "skip", self._section_type not in DEFAULT_NOSKIP
                ):
                    continue

            if self._active_section is None:
                continue

            active_section = self._active_section
            section_type = self._section_type
            section_name = self._section_name

            if active_section.get(
                "use_unmatched",
                self._section_type not in DEFAULT_NOT_USE_UNMATCHED,
            ):
                line = line_unmatched.strip()
                if len(line) == 0:
                    continue

            if section_type == "table":
                sep = active_section.get("separator", ",")
                row = line.split(sep)

                if section_name not in output:
                    # Table needs initialisation
                    (
                        has_header,
                        row,
                        table_header,
                        column_types,
                    ) = self._parse_table(active_section, row)

                    output[section_name] = {k: [] for k in table_header}

                    if active_section.get("has_header", True):
                        continue

                if len(row) < len(table_header):
                    # skip lines that have fewer columns than expected
                    continue

                # Merge extra columns into final column

                row = self._table_merge_extra_columns(
                    table_header, sep, row, column_types
                )

                # Fill out current row
                for val, colname, coltype in zip(
                    row, table_header, column_types
                ):
                    output[section_name][colname].append(
                        _map_to_type(val.strip(), coltype)
                    )

            elif section_type in {"list", "lists"}:
                sep = active_section.get("separator", ",")
                output[section_name] = output.get(section_name, [])

                map_type = active_section.get("map")
                next_list = [
                    _map_to_type(el.strip(), map_type)
                    for el in line.split(sep)
                ]

                list_to_append = (
                    [next_list] if section_type == "lists" else next_list
                )
                output[section_name] += list_to_append

            elif section_type in {"regex", "regexs"}:
                regex = active_section.get("regex", "^(.*)$")
                map_type = active_section.get("map")

                matches = re.findall(regex, line)
                if len(matches) == 0:
                    continue
                elif len(matches) == 1 and section_type == "regex":
                    output[section_name] = _map_to_type(matches[0], map_type)
                else:
                    output[section_name] = output.get(section_name, [])
                    output[section_name] += [
                        _map_to_type(m, map_type) for m in matches
                    ]

                # Terminate after finding the first match
                self._terminate_after_first_match(active_section, section_type)

            elif section_type == "stop":
                break

            else:
                # By default, just append additional lines as text
                new_str = (
                    f"{output[section_name]}\n{line}"
                    if section_name in output
                    else line
                )
                output[section_name] = new_str

        return output

    @staticmethod
    def determine_unmatched_part(
        matches: t.Dict[str, t.List], line_pp: t.List[str]
    ):

        if matches["startswith"]:
            _, t = matches["startswith"][0]
            line_unmatched = [
                line[len(t) :] for line in line_pp if line.startswith(t)
            ][0]
        elif matches["endswith"]:
            _, t = matches["endwith"][0]
            line_unmatched = [
                line[: -(len(t) + 1)] for line in line_pp if line.endswith(t)
            ][0]
        elif matches["contains"]:
            _, t = matches["contains"][0]
            lpp = [line for line in line_pp if line.find(t) >= 0][0]
            i = lpp.find(t)
            line_unmatched = lpp[:i] + lpp[(i + len(t)) :]
        elif matches["re"]:
            _, r = matches["re"][0]
            line_unmatched = [
                r.sub("", line) for line in line_pp if len(r.findall(line)) > 0
            ][0]
        return line_unmatched

    def _terminate_after_first_match(self, active_section, section_type):
        # Terminate after finding the first match
        if section_type == "regex":
            next_section = active_section.get("next_section")
            self._set_section(next_section)
        return next_section

    @staticmethod
    def _parse_table(active_section, row):

        has_header = active_section.get("has_header", True)
        if has_header:
            row = [col.strip() for col in row]
        default_type = active_section.get("default_map", "str")
        colmap = active_section.get("column_map", len(row) * [(None, None)])
        if type(colmap) == list:
            # Columns are defined in order
            if has_header:
                table_header = [mn or rn for rn, (mn, _) in zip(row, colmap)]
                table_header += row[len(colmap) :]
                column_types = [mt for _, mt in colmap]
                column_types += (len(row) - len(colmap)) * [default_type]
            else:
                table_header = [
                    mn or "column{:02d}".format(i + 1)
                    for i, (mn, _) in enumerate(colmap)
                ]
                column_types = [mt or default_type for _, mt in colmap]
        elif type(colmap) == dict:
            if not has_header:
                raise ParseError("dict column maps must have a header")
            # First row is a header
            table_header = [colmap.get(rn, (rn, None))[0] for rn in row]
            column_types = [
                colmap.get(rn, (None, default_type))[1] for rn in row
            ]
        else:
            raise ParseError("badly formatted column map")
        return has_header, row, table_header, column_types

    @staticmethod
    def _table_merge_extra_columns(table_header, sep, row, column_types):
        # Merge extra columns into final column
        ncol = len(table_header)
        if len(row) > ncol:
            row[ncol - 1] = sep.join(row[ncol - 1 :])
            del row[ncol:]
        assert len(row) == len(table_header) and len(row) == len(column_types)
        return row


def _map_to_type(val, map_type):
    if map_type and map_type.startswith("datetime"):
        date_format = "%Y-%m-%dT%H:%M:%S.%fZ"  # ISO 8601 format
        if map_type.startswith("datetime:"):
            date_format = map_type[9:]
        try:
            return datetime.strptime(val, date_format)
        except ValueError:
            return None
    else:
        try:
            return {"str": str, "int": int, "float": float, "bool": bool}.get(
                map_type, str
            )(val)
        except ValueError or TypeError:
            return {"float": float("nan")}.get(map_type)



import logging
import typing as t
from copy import copy
from functools import cached_property, lru_cache
from pathlib import PosixPath

import bottleneck as bn
import h5py
import numpy as np
import pandas as pd

from agora.io.bridge import BridgeH5
from agora.io.decorators import _first_arg_str_to_df
from agora.utils.association import validate_association
from agora.utils.kymograph import add_index_levels
from agora.utils.merge import apply_merges


class Signal(BridgeH5):
    """
    Fetch data from h5 files for post-processing.

    Signal assumes that the metadata and data are accessible to perform time-adjustments and apply previously recorded post-processes.
    """

    def __init__(self, file: t.Union[str, PosixPath]):
        """Define index_names for dataframes, candidate fluorescence channels, and composite statistics."""
        super().__init__(file, flag=None)
        self.index_names = (
            "experiment",
            "position",
            "trap",
            "cell_label",
            "mother_label",
        )
        self.candidate_channels = (
            "GFP",
            "GFPFast",
            "mCherry",
            "Flavin",
            "Citrine",
            "mKO2",
            "Cy5",
            "pHluorin405",
        )

    def __getitem__(self, dsets: t.Union[str, t.Collection]):
        """Get and potentially pre-process data from h5 file and return as a dataframe."""
        if isinstance(dsets, str):  # no pre-processing
            df = self.get_raw(dsets)
            return self.add_name(df, dsets)
        elif isinstance(dsets, list):  # pre-processing
            is_bgd = [dset.endswith("imBackground") for dset in dsets]
            # Check we are not comparing tile-indexed and cell-indexed data
            assert sum(is_bgd) == 0 or sum(is_bgd) == len(
                dsets
            ), "Tile data and cell data can't be mixed"
            return [
                self.add_name(self.apply_prepost(dset), dset) for dset in dsets
            ]
        else:
            raise Exception(f"Invalid type {type(dsets)} to get datasets")

    @staticmethod
    def add_name(df, name):
        """Add column of identical strings to a dataframe."""
        df.name = name
        return df

    def cols_in_mins(self, df: pd.DataFrame):
        # Convert numerical columns in a dataframe to minutes
        try:
            df.columns = (df.columns * self.tinterval // 60).astype(int)
        except Exception as e:
            self._log(f"Unable to convert columns to minutes: {e}", "debug")
        return df

    @cached_property
    def ntimepoints(self):
        """Find the number of time points for one position, or one h5 file."""
        with h5py.File(self.filename, "r") as f:
            return f["extraction/general/None/area/timepoint"][-1] + 1

    @cached_property
    def tinterval(self) -> int:
        """Find the interval between time points (minutes)."""
        tinterval_location = "time_settings/timeinterval"
        with h5py.File(self.filename, "r") as f:
            if tinterval_location in f.attrs:
                return f.attrs[tinterval_location][0]
            else:
                logging.getlogger("aliby").warn(
                    f"{str(self.filename).split('/')[-1]}: using default time interval of 5 minutes"
                )
                return 5

    @staticmethod
    def get_retained(df, cutoff):
        """Return a fraction of the df, one without later time points."""
        return df.loc[bn.nansum(df.notna(), axis=1) > df.shape[1] * cutoff]

    @property
    def channels(self) -> t.Collection[str]:
        """Get channels as an array of strings."""
        with h5py.File(self.filename, "r") as f:
            return list(f.attrs["channels"])

    @_first_arg_str_to_df
    def retained(self, signal, cutoff=0.8):
        """
        Load data (via decorator) and reduce the resulting dataframe.

        Load data for a signal or a list of signals and reduce the resulting
        dataframes to a fraction of their original size, losing late time
        points.
        """
        if isinstance(signal, pd.DataFrame):
            return self.get_retained(signal, cutoff)
        elif isinstance(signal, list):
            return [self.get_retained(d, cutoff=cutoff) for d in signal]

    @lru_cache(2)
    def lineage(
        self, lineage_location: t.Optional[str] = None, merged: bool = False
    ) -> np.ndarray:
        """
        Get lineage data from a given location in the h5 file.

        Returns an array with three columns: the tile id, the mother label, and the daughter label.
        """
        if lineage_location is None:
            lineage_location = "postprocessing/lineage"
            if merged:
                lineage_location += "_merged"
        with h5py.File(self.filename, "r") as f:
            tile_mo_da = f[lineage_location]
            lineage = np.array(
                (
                    tile_mo_da["trap"],
                    tile_mo_da["mother_label"],
                    tile_mo_da["daughter_label"],
                )
            ).T
        return lineage

    @_first_arg_str_to_df
    def apply_prepost(
        self,
        data: t.Union[str, pd.DataFrame],
        merges: t.Union[np.ndarray, bool] = True,
        picks: t.Union[t.Collection, bool] = True,
    ):
        """
        Apply modifier operations (picker or merger) to a dataframe.

        Parameters
        ----------
        data : t.Union[str, pd.DataFrame]
            DataFrame or path to one.
        merges : t.Union[np.ndarray, bool]
            (optional) 2-D array with three columns: the tile id, the mother label, and the daughter id.
            If True, fetch merges from file.
        picks : t.Union[np.ndarray, bool]
            (optional) 2-D array with two columns: the tiles and
            the cell labels.
            If True, fetch picks from file.

        Examples
        --------
        FIXME: Add docs.

        """
        if isinstance(merges, bool):
            merges: np.ndarray = self.get_merges() if merges else np.array([])
        if merges.any():
            merged = apply_merges(data, merges)
        else:
            merged = copy(data)
        if isinstance(picks, bool):
            picks = (
                self.get_picks(names=merged.index.names)
                if picks
                else set(merged.index)
            )
        with h5py.File(self.filename, "r") as f:
            if "modifiers/picks" in f and picks:
                if picks:
                    return merged.loc[
                        set(picks).intersection(
                            [tuple(x) for x in merged.index]
                        )
                    ]
                else:
                    if isinstance(merged.index, pd.MultiIndex):
                        empty_lvls = [[] for i in merged.index.names]
                        index = pd.MultiIndex(
                            levels=empty_lvls,
                            codes=empty_lvls,
                            names=merged.index.names,
                        )
                    else:
                        index = pd.Index([], name=merged.index.name)
                    merged = pd.DataFrame([], index=index)
        return merged

    @cached_property
    def p_available(self):
        """Print data sets available in h5 file."""
        if not hasattr(self, "_available"):
            self._available = []
            with h5py.File(self.filename, "r") as f:
                f.visititems(self.store_signal_path)
        for sig in self._available:
            print(sig)

    @cached_property
    def available(self):
        """Get data sets available in h5 file."""
        try:
            if not hasattr(self, "_available"):
                self._available = []
            with h5py.File(self.filename, "r") as f:
                f.visititems(self.store_signal_path)
        except Exception as e:
            self._log("Exception when visiting h5: {}".format(e), "exception")

        return self._available

    def get_merged(self, dataset):
        """Run preprocessing for merges."""
        return self.apply_prepost(dataset, picks=False)

    @cached_property
    def merges(self) -> np.ndarray:
        """Get merges."""
        with h5py.File(self.filename, "r") as f:
            dsets = f.visititems(self._if_merges)
        return dsets

    @cached_property
    def n_merges(self):
        """Get number of merges."""
        return len(self.merges)

    @cached_property
    def picks(self) -> np.ndarray:
        """Get picks."""
        with h5py.File(self.filename, "r") as f:
            dsets = f.visititems(self._if_picks)
        return dsets

    def get_raw(
        self,
        dataset: str or t.List[str],
        in_minutes: bool = True,
        lineage: bool = False,
    ) -> pd.DataFrame or t.List[pd.DataFrame]:
        """
        Load data from a h5 file and return as a dataframe.

        Parameters
        ----------
        dataset: str or list of strs
            The name of the h5 file or a list of h5 file names
        in_minutes: boolean
            If True,
        lineage: boolean
        """
        try:
            if isinstance(dataset, str):
                with h5py.File(self.filename, "r") as f:
                    df = self.dataset_to_df(f, dataset).sort_index()
                    if in_minutes:
                        df = self.cols_in_mins(df)
            elif isinstance(dataset, list):
                return [
                    self.get_raw(dset, in_minutes=in_minutes, lineage=lineage)
                    for dset in dataset
                ]
            if lineage:  # assume that df is sorted
                mother_label = np.zeros(len(df), dtype=int)
                lineage = self.lineage()
                a, b = validate_association(
                    lineage,
                    np.array(df.index.to_list()),
                    match_column=1,
                )
                mother_label[b] = lineage[a, 1]
                df = add_index_levels(df, {"mother_label": mother_label})
            return df
        except Exception as e:
            self._log(f"Could not fetch dataset {dataset}: {e}", "error")
            raise e

    def get_merges(self):
        """Get merge events going up to the first level."""
        with h5py.File(self.filename, "r") as f:
            merges = f.get("modifiers/merges", np.array([]))
            if not isinstance(merges, np.ndarray):
                merges = merges[()]
        return merges

    def get_picks(
        self,
        names: t.Tuple[str, ...] = ("trap", "cell_label"),
        path: str = "modifiers/picks/",
    ) -> t.Set[t.Tuple[int, str]]:
        """Get the relevant picks based on names."""
        with h5py.File(self.filename, "r") as f:
            picks = set()
            if path in f:
                picks = set(zip(*[f[path + name] for name in names]))
            return picks

    def dataset_to_df(self, f: h5py.File, path: str) -> pd.DataFrame:
        """Get data from h5 file as a dataframe."""
        assert path in f, f"{path} not in {f}"
        dset = f[path]
        values, index, columns = [], [], []
        index_names = copy(self.index_names)
        valid_names = [lbl for lbl in index_names if lbl in dset.keys()]
        if valid_names:
            index = pd.MultiIndex.from_arrays(
                [dset[lbl] for lbl in valid_names], names=valid_names
            )
            columns = dset.attrs.get("columns", None)
            if "timepoint" in dset:
                columns = f[path + "/timepoint"][()]
            values = f[path + "/values"][()]
        df = pd.DataFrame(values, index=index, columns=columns)
        return df

    @property
    def stem(self):
        """Get name of h5 file."""
        return self.filename.stem

    def store_signal_path(
        self,
        fullname: str,
        node: t.Union[h5py.Dataset, h5py.Group],
    ):
        """
        Store the name of a signal if it is a leaf node
        (a group with no more groups inside) and if it starts with extraction.
        """
        if isinstance(node, h5py.Group) and np.all(
            [isinstance(x, h5py.Dataset) for x in node.values()]
        ):
            self._if_ext_or_post(fullname, self._available)

    @staticmethod
    def _if_ext_or_post(name: str, siglist: list):
        if name.startswith("extraction") or name.startswith("postprocessing"):
            siglist.append(name)

    @staticmethod
    def _if_merges(name: str, obj):
        if isinstance(obj, h5py.Dataset) and name.startswith(
            "modifiers/merges"
        ):
            return obj[()]

    @staticmethod
    def _if_picks(name: str, obj):
        if isinstance(obj, h5py.Group) and name.endswith("picks"):
            return obj[()]

    # TODO FUTURE add stages support to fluigent system
    @property
    def ntps(self) -> int:
        """Get number of time points from the metadata."""
        return self.meta_h5["time_settings/ntimepoints"][0]

    @property
    def stages(self) -> t.List[str]:
        """Get the contents of the pump with highest flow rate at each stage."""
        flowrate_name = "pumpinit/flowrate"
        pumprate_name = "pumprate"
        switchtimes_name = "switchtimes"
        main_pump_id = np.concatenate(
            (
                (np.argmax(self.meta_h5[flowrate_name]),),
                np.argmax(self.meta_h5[pumprate_name], axis=0),
            )
        )
        if not self.meta_h5[switchtimes_name][0]:  # Cover for t0 switches
            main_pump_id = main_pump_id[1:]
        return [self.meta_h5["pumpinit/contents"][i] for i in main_pump_id]

    @property
    def nstages(self) -> int:
        return len(self.switch_times) + 1

    @property
    def max_span(self) -> int:
        return int(self.tinterval * self.ntps / 60)

    @property
    def switch_times(self) -> t.List[int]:
        switchtimes_name = "switchtimes"
        switches_minutes = self.meta_h5[switchtimes_name]
        return [
            t_min
            for t_min in switches_minutes
            if t_min and t_min < self.max_span
        ]  # Cover for t0 switches

    @property
    def stages_span(self) -> t.Tuple[t.Tuple[str, int], ...]:
        """Get consecutive stages and their corresponding number of time points."""
        transition_tps = (0, *self.switch_times, self.max_span)
        spans = [
            end - start
            for start, end in zip(transition_tps[:-1], transition_tps[1:])
            if end <= self.max_span
        ]
        return tuple((stage, ntps) for stage, ntps in zip(self.stages, spans))

    @property
    def stages_span_tp(self) -> t.Tuple[t.Tuple[str, int], ...]:
        return tuple(
            [
                (name, (t_min * 60) // self.tinterval)
                for name, t_min in self.stages_span
            ]
        )



"""
Tools to manage I/O using a remote OMERO server.
"""

import re
import typing as t
from abc import abstractmethod
from pathlib import PosixPath

import dask.array as da
import numpy as np
import omero
from dask import delayed
from omero.gateway import BlitzGateway
from omero.model import enums as omero_enums
from yaml import safe_load

from agora.io.bridge import BridgeH5

# convert OMERO definitions into numpy types
PIXEL_TYPES = {
    omero_enums.PixelsTypeint8: np.int8,
    omero_enums.PixelsTypeuint8: np.uint8,
    omero_enums.PixelsTypeint16: np.int16,
    omero_enums.PixelsTypeuint16: np.uint16,
    omero_enums.PixelsTypeint32: np.int32,
    omero_enums.PixelsTypeuint32: np.uint32,
    omero_enums.PixelsTypefloat: np.float32,
    omero_enums.PixelsTypedouble: np.float64,
}


class BridgeOmero:
    """
    Core to interact with OMERO, using credentials or fetching them from h5 file (temporary trick).
    See
    https://docs.openmicroscopy.org/omero/5.6.0/developers/Python.html
    """

    def __init__(
        self,
        host: str = None,
        username: str = None,
        password: str = None,
        ome_id: int = None,
    ):
        """
        Parameters
        ----------
        host : string
            web address of OMERO host
        username: string
        password : string
        ome_id: Optional int
            Unique identifier on Omero database. Used to fetch specific objects.
        """
        # assert all((host, username, password)), str(f"Invalid credentials host:{host}, user:{username}, pass:{pass}")
        assert all(
            (host, username, password)
        ), f"Invalid credentials. host: {host}, user: {username}, pwd: {password}"

        self.conn = None
        self.host = host
        self.username = username
        self.password = password
        self.ome_id = ome_id

    # standard method required for Python's with statement
    def __enter__(self):
        self.create_gate()

        return self

    @property
    def ome_class(self):
        # Initialise Omero Object Wrapper for instances when applicable.
        if not hasattr(self, "_ome_class"):
            assert (
                self.conn.isConnected() and self.ome_id is not None
            ), "No Blitz connection or valid omero id"

            ome_type = [
                valid_name
                for valid_name in ("Dataset", "Image")
                if re.match(
                    f".*{ valid_name }.*",
                    self.__class__.__name__,
                    re.IGNORECASE,
                )
            ][0]
            self._ome_class = self.conn.getObject(ome_type, self.ome_id)

            assert self._ome_class, f"{ome_type} {self.ome_id} not found."

        return self._ome_class

    def create_gate(self) -> bool:
        self.conn = BlitzGateway(
            host=self.host, username=self.username, passwd=self.password
        )
        self.conn.connect()
        self.conn.c.enableKeepAlive(60)

        self.conn.isConnected()

    # standard method required for Python's with statement
    def __exit__(self, *exc) -> bool:
        for e in exc:
            if e is not None:
                print(e)

        self.conn.close()
        return False

    @classmethod
    def server_info_from_h5(
        cls,
        filepath: t.Union[str, PosixPath],
    ):
        """Return server info from hdf5 file.

        Parameters
        ----------
        cls : BridgeOmero
            BridgeOmero class
        filepath : t.Union[str, PosixPath]
            Location of hdf5 file.

        Examples
        --------
        FIXME: Add docs.

        """
        # metadata = load_attributes(filepath)
        bridge = BridgeH5(filepath)
        meta = safe_load(bridge.meta_h5["parameters"])["general"]
        server_info = {k: meta[k] for k in ("host", "username", "password")}
        return server_info

    def set_id(self, ome_id: int):
        self.ome_id = ome_id

    @property
    def file_annotations(self):
        valid_annotations = [
            ann.getFileName()
            for ann in self.ome_class.listAnnotations()
            if hasattr(ann, "getFileName")
        ]
        return valid_annotations

    def add_file_as_annotation(
        self, file_to_upload: t.Union[str, PosixPath], **kwargs
    ):
        """Upload annotation to object on OMERO server. Only valid in subclasses.

        Parameters
        ----------
        file_to_upload: File to upload
        **kwargs: Additional keyword arguments passed on
            to BlitzGateway.createFileAnnfromLocalFile
        """

        file_annotation = self.conn.createFileAnnfromLocalFile(
            file_to_upload,
            mimetype="text/plain",
            **kwargs,
        )
        self.ome_class.linkAnnotation(file_annotation)


class Dataset(BridgeOmero):
    """
    Tool to interact with Omero Datasets remotely, access their
    metadata and associated files and images.


    Parameters
    ----------
    expt_id: int Dataset id on server
    server_info: dict host, username and password

    """

    def __init__(self, expt_id: int, **server_info):
        super().__init__(ome_id=expt_id, **server_info)

    @property
    def name(self):
        return self.ome_class.getName()

    @property
    def date(self):
        return self.ome_class.getDate()

    @property
    def unique_name(self):
        return "_".join(
            (
                str(self.ome_id),
                self.date.strftime("%Y_%m_%d").replace("/", "_"),
                self.name,
            )
        )

    def get_images(self):
        return {
            im.getName(): im.getId() for im in self.ome_class.listChildren()
        }

    @property
    def files(self):
        if not hasattr(self, "_files"):
            self._files = {
                x.getFileName(): x
                for x in self.ome_class.listAnnotations()
                if isinstance(x, omero.gateway.FileAnnotationWrapper)
            }
        if not len(self._files):
            raise Exception(
                "exception:metadata: experiment has no annotation files."
            )
        elif len(self.file_annotations) != len(self._files):
            raise Exception("Number of files and annotations do not match")

        return self._files

    @property
    def tags(self):
        if self._tags is None:
            self._tags = {
                x.getname(): x
                for x in self.ome_class.listAnnotations()
                if isinstance(x, omero.gateway.TagAnnotationWrapper)
            }
        return self._tags

    def cache_logs(self, root_dir):
        valid_suffixes = ("txt", "log")
        for _, annotation in self.files.items():
            filepath = root_dir / annotation.getFileName().replace("/", "_")
            if (
                any([str(filepath).endswith(suff) for suff in valid_suffixes])
                and not filepath.exists()
            ):
                # save only the text files
                with open(str(filepath), "wb") as fd:
                    for chunk in annotation.getFileInChunks():
                        fd.write(chunk)
        return True

    @classmethod
    def from_h5(
        cls,
        filepath: t.Union[str, PosixPath],
    ):
        """Instatiate Dataset from a hdf5 file.

        Parameters
        ----------
        cls : Image
            Image class
        filepath : t.Union[str, PosixPath]
            Location of hdf5 file.

        Examples
        --------
        FIXME: Add docs.

        """
        # metadata = load_attributes(filepath)
        bridge = BridgeH5(filepath)
        dataset_keys = ("omero_id", "omero_id,", "dataset_id")
        for k in dataset_keys:
            if k in bridge.meta_h5:
                return cls(
                    bridge.meta_h5[k], **cls.server_info_from_h5(filepath)
                )


class Image(BridgeOmero):
    """
    Loads images from OMERO and gives access to the data and metadata.
    """

    def __init__(self, image_id: int, **server_info):
        """
        Establishes the connection to the OMERO server via the Argo
        base class.

        Parameters
        ----------
        image_id: integer
        server_info: dictionary
            Specifies the host, username, and password as strings
        """
        super().__init__(ome_id=image_id, **server_info)

    @classmethod
    def from_h5(
        cls,
        filepath: t.Union[str, PosixPath],
    ):
        """Instatiate Image from a hdf5 file.

        Parameters
        ----------
        cls : Image
            Image class
        filepath : t.Union[str, PosixPath]
            Location of hdf5 file.

        Examples
        --------
        FIXME: Add docs.

        """
        # metadata = load_attributes(filepath)
        bridge = BridgeH5(filepath)
        image_id = bridge.meta_h5["image_id"]
        return cls(image_id, **cls.server_info_from_h5(filepath))

    @property
    def name(self):
        return self.ome_class.getName()

    @property
    def data(self):
        return get_data_lazy(self.ome_class)

    @property
    def metadata(self):
        """
        Store metadata saved in OMERO: image size, number of time points,
        labels of channels, and image name.
        """
        meta = dict()
        meta["size_x"] = self.ome_class.getSizeX()
        meta["size_y"] = self.ome_class.getSizeY()
        meta["size_z"] = self.ome_class.getSizeZ()
        meta["size_c"] = self.ome_class.getSizeC()
        meta["size_t"] = self.ome_class.getSizeT()
        meta["channels"] = self.ome_class.getChannelLabels()
        meta["name"] = self.ome_class.getName()
        return meta


class UnsafeImage(Image):
    """
    Loads images from OMERO and gives access to the data and metadata.
    This class is a temporary solution while we find a way to use
    context managers inside napari. It risks resulting in zombie connections
    and producing freezes in an OMERO server.

    """

    def __init__(self, image_id, **server_info):
        """
        Establishes the connection to the OMERO server via the Argo
        base class.

        Parameters
        ----------
        image_id: integer
        server_info: dictionary
            Specifies the host, username, and password as strings
        """
        super().__init__(image_id, **server_info)
        self.create_gate()

    @property
    def data(self):
        try:
            return get_data_lazy(self.ome_class)
        except Exception as e:
            print(f"ERROR: Failed fetching image from server: {e}")
            self.conn.connect(False)


def get_data_lazy(image) -> da.Array:
    """
    Get 5D dask array, with delayed reading from OMERO image.
    """
    nt, nc, nz, ny, nx = [getattr(image, f"getSize{x}")() for x in "TCZYX"]
    pixels = image.getPrimaryPixels()
    dtype = PIXEL_TYPES.get(pixels.getPixelsType().value, None)
    # using dask
    get_plane = delayed(lambda idx: pixels.getPlane(*idx))

    def get_lazy_plane(zct):
        return da.from_delayed(get_plane(zct), shape=(ny, nx), dtype=dtype)

    # 5D stack: TCZXY
    t_stacks = []
    for t in range(nt):
        c_stacks = []
        for c in range(nc):
            z_stack = []
            for z in range(nz):
                z_stack.append(get_lazy_plane((z, c, t)))
            c_stacks.append(da.stack(z_stack))
        t_stacks.append(da.stack(c_stacks))

    return da.stack(t_stacks)



import logging
import typing as t
from itertools import groupby
from pathlib import Path, PosixPath
from functools import lru_cache, cached_property

import h5py
import numpy as np
from numpy.lib.stride_tricks import sliding_window_view
from scipy import ndimage
from scipy.sparse.base import isdense
from utils_find_1st import cmp_equal, find_1st


class Cells:
    """
    Extracts information from an h5 file. This class accesses:

    'cell_info', which contains 'angles', 'cell_label', 'centres',
    'edgemasks', 'ellipse_dims', 'mother_assign', 'mother_assign_dynamic',
    'radii', 'timepoint', 'trap'.
    All of these except for 'edgemasks' are a 1D ndarray.

    'trap_info', which contains 'drifts', 'trap_locations'

    """

    def __init__(self, filename, path="cell_info"):
        self.filename: t.Optional[t.Union[str, PosixPath]] = filename
        self.cinfo_path: t.Optional[str] = path
        self._edgemasks: t.Optional[str] = None
        self._tile_size: t.Optional[int] = None

    @classmethod
    def from_source(cls, source: t.Union[PosixPath, str]):
        return cls(Path(source))

    def _log(self, message: str, level: str = "warn"):
        # Log messages in the corresponding level
        logger = logging.getLogger("aliby")
        getattr(logger, level)(f"{self.__class__.__name__}: {message}")

    @staticmethod
    def _asdense(array: np.ndarray):
        if not isdense(array):
            array = array.todense()
        return array

    @staticmethod
    def _astype(array: np.ndarray, kind: str):
        # Convert sparse arrays if needed and if kind is 'mask' it fills the outline
        array = Cells._asdense(array)
        if kind == "mask":
            array = ndimage.binary_fill_holes(array).astype(bool)
        return array

    def _get_idx(self, cell_id: int, trap_id: int):
        # returns boolean array of time points where both the cell with cell_id and the trap with trap_id exist
        return (self["cell_label"] == cell_id) & (self["trap"] == trap_id)

    @property
    def max_labels(self) -> t.List[int]:
        return [max((0, *self.labels_in_trap(i))) for i in range(self.ntraps)]

    @property
    def max_label(self) -> int:
        return sum(self.max_labels)

    @property
    def ntraps(self) -> int:
        # find the number of traps from the h5 file
        with h5py.File(self.filename, mode="r") as f:
            return len(f["trap_info/trap_locations"][()])

    @property
    def tinterval(self):
        with h5py.File(self.filename, mode="r") as f:
            return f.attrs["time_settings/timeinterval"]

    @property
    def traps(self) -> t.List[int]:
        # returns a list of traps
        return list(set(self["trap"]))

    @property
    def tile_size(self) -> t.Union[int, t.Tuple[int], None]:
        if self._tile_size is None:
            with h5py.File(self.filename, mode="r") as f:
                # self._tile_size = f["trap_info/tile_size"][0]
                self._tile_size = f["cell_info/edgemasks"].shape[1:]
        return self._tile_size

    def nonempty_tp_in_trap(self, trap_id: int) -> set:
        # given a trap_id returns time points in which cells are available
        return set(self["timepoint"][self["trap"] == trap_id])

    @property
    def edgemasks(self) -> t.List[np.ndarray]:
        # returns the masks per tile
        if self._edgemasks is None:
            edgem_path: str = "edgemasks"
            self._edgemasks = self._fetch(edgem_path)
        return self._edgemasks

    @property
    def labels(self) -> t.List[t.List[int]]:
        """
        Return all cell labels in object
        We use mother_assign to list traps because it is the only property that appears even
        when no cells are found
        """
        return [self.labels_in_trap(trap) for trap in range(self.ntraps)]

    def max_labels_in_frame(self, frame: int) -> t.List[int]:
        # Return the maximum label for each trap in the given frame
        max_labels = [
            self["cell_label"][
                (self["timepoint"] <= frame) & (self["trap"] == trap_id)
            ]
            for trap_id in range(self.ntraps)
        ]
        return [max([0, *labels]) for labels in max_labels]

    def where(self, cell_id: int, trap_id: int):
        """
        Parameters
        ----------
            cell_id: int
                Cell index
            trap_id: int
                Trap index

        Returns
        ----------
            indices int array
            boolean mask array
            edge_ix int array
        """
        indices = self._get_idx(cell_id, trap_id)
        edgem_ix = self._edgem_where(cell_id, trap_id)
        return (
            self["timepoint"][indices],
            indices,
            edgem_ix,
        )

    def mask(self, cell_id, trap_id):
        """
        Returns the times and the binary masks of a given cell in a given tile.

        Parameters
        ----------
        cell_id : int
            The unique ID of the cell.
        tile_id : int
            The unique ID of the tile.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            The times when the binary masks were taken and the binary masks of the given cell in the given tile.

        """
        times, outlines = self.outline(cell_id, trap_id)
        return times, np.array(
            [ndimage.morphology.binary_fill_holes(o) for o in outlines]
        )

    def at_time(
        self, timepoint: t.Iterable[int], kind="mask"
    ) -> t.List[t.List[np.ndarray]]:
        """
        Returns a list of lists of binary masks in a given list of time points.

        Parameters
        ----------
        timepoints : Iterable[int]
            The list of time points for which to return the binary masks.
        kind : str, optional
            The type of binary masks to return, by default "mask".

        Returns
        -------
        List[List[np.ndarray]]
            A list of lists with binary masks grouped by tile IDs.

        """

        ix = self["timepoint"] == timepoint
        traps = self["trap"][ix]
        edgemasks = self._edgem_from_masking(ix)
        masks = [
            self._astype(edgemask, kind)
            for edgemask in edgemasks
            if edgemask.any()
        ]
        return self.group_by_traps(traps, masks)

    def at_times(
        self, timepoints: t.Iterable[int], kind="mask"
    ) -> t.List[t.List[np.ndarray]]:
        """
        Returns a list of lists of binary masks for a given list of time points.

        Parameters
        ----------
        timepoints : Iterable[int]
            The list of time points for which to return the binary masks.
        kind : str, optional
            The type of binary masks to return, by default "mask".

        Returns
        -------
        List[List[np.ndarray]]
            A list of lists with binary masks grouped by tile IDs.

        """
        return [
            [
                np.stack(tile_masks) if len(tile_masks) else []
                for tile_masks in self.at_time(tp, kind=kind).values()
            ]
            for tp in timepoints
        ]

    def group_by_traps(
        self, traps: t.Collection, cell_labels: t.Collection
    ) -> t.Dict[int, t.List[int]]:
        """
        Returns a dict with traps as keys and list of labels as value.
        Note that the total number of traps are calculated from Cells.traps.

        """
        iterator = groupby(zip(traps, cell_labels), lambda x: x[0])
        d = {key: [x[1] for x in group] for key, group in iterator}
        d = {i: d.get(i, []) for i in self.traps}
        return d

    def labels_in_trap(self, trap_id: int) -> t.Set[int]:
        # return set of cell ids for a given trap
        return set((self["cell_label"][self["trap"] == trap_id]))

    def labels_at_time(self, timepoint: int) -> t.Dict[int, t.List[int]]:
        labels = self["cell_label"][self["timepoint"] == timepoint]
        traps = self["trap"][self["timepoint"] == timepoint]
        return self.group_by_traps(traps, labels)

    def __getitem__(self, item):
        assert item != "edgemasks", "Edgemasks must not be loaded as a whole"

        _item = "_" + item
        if not hasattr(self, _item):
            setattr(self, _item, self._fetch(item))
        return getattr(self, _item)

    def _fetch(self, path):
        with h5py.File(self.filename, mode="r") as f:
            return f[self.cinfo_path][path][()]

    def _edgem_from_masking(self, mask):
        with h5py.File(self.filename, mode="r") as f:
            edgem = f[self.cinfo_path + "/edgemasks"][mask, ...]
        return edgem

    def _edgem_where(self, cell_id, trap_id):
        id_mask = self._get_idx(cell_id, trap_id)
        edgem = self._edgem_from_masking(id_mask)

        return edgem

    def outline(self, cell_id: int, trap_id: int):
        id_mask = self._get_idx(cell_id, trap_id)
        times = self["timepoint"][id_mask]

        return times, self._edgem_from_masking(id_mask)

    @property
    def ntimepoints(self) -> int:
        return self["timepoint"].max() + 1

    @cached_property
    def _cells_vs_tps(self):
        # Binary matrix showing the presence of all cells in all time points
        ncells_per_tile = [len(x) for x in self.labels]
        cells_vs_tps = np.zeros(
            (sum(ncells_per_tile), self.ntimepoints), dtype=bool
        )

        cells_vs_tps[
            self._cell_cumsum[self["trap"]] + self["cell_label"] - 1,
            self["timepoint"],
        ] = True
        return cells_vs_tps

    @cached_property
    def _cell_cumsum(self):
        # Cumulative sum indicating the number of cells per tile
        ncells_per_tile = [len(x) for x in self.labels]
        cumsum = np.roll(np.cumsum(ncells_per_tile), shift=1)
        cumsum[0] = 0
        return cumsum

    def _flat_index_to_tuple_location(self, idx: int) -> t.Tuple[int, int]:
        # Convert a cell index to a tuple
        # Note that it assumes tiles and cell labels are flattened, but
        # it is agnostic to tps

        tile_id = int(np.where(idx + 1 > self._cell_cumsum)[0][-1])
        cell_label = idx - self._cell_cumsum[tile_id] + 1
        return tile_id, cell_label

    @property
    def _tiles_vs_cells_vs_tps(self):
        ncells_mat = np.zeros(
            (self.ntraps, self["cell_label"].max(), self.ntimepoints),
            dtype=bool,
        )
        ncells_mat[
            self["trap"], self["cell_label"] - 1, self["timepoint"]
        ] = True
        return ncells_mat

    def cell_tp_where(
        self,
        min_consecutive_tps: int = 15,
        interval: None or t.Tuple[int, int] = None,
    ):
        window = sliding_window_view(
            self._cells_vs_tps, min_consecutive_tps, axis=1
        )

        tp_min = window.sum(axis=-1) == min_consecutive_tps

        # Apply an interval filter to focucs on a slice
        if interval is not None:
            interval = tuple(np.array(interval))
        else:
            interval = (0, window.shape[1])

        low_boundary, high_boundary = interval

        tp_min[:, :low_boundary] = False
        tp_min[:, high_boundary:] = False
        return tp_min

    @lru_cache(20)
    def mothers_in_trap(self, trap_id: int):
        return self.mothers[trap_id]

    @cached_property
    def mothers(self):
        """
        Return nested list with final prediction of mother id for each cell
        """
        return self.mother_assign_from_dynamic(
            self["mother_assign_dynamic"],
            self["cell_label"],
            self["trap"],
            self.ntraps,
        )

    @cached_property
    def mothers_daughters(self) -> np.ndarray:
        """
        Return a single array with three columns, containing information about
        the mother-daughter relationships: tile, mothers and daughters.

        Returns
        -------
        np.ndarray
            An array with shape (n, 3) where n is the number of mother-daughter pairs found.
            The columns contain:
            - tile: the tile where the mother cell is located.
            - mothers: the index of the mother cell within the tile.
            - daughters: the index of the daughter cell within the tile.
        """
        nested_massign = self.mothers

        if sum([x for y in nested_massign for x in y]):
            mothers_daughters = np.array(
                [
                    (tid, m, d)
                    for tid, trapcells in enumerate(nested_massign)
                    for d, m in enumerate(trapcells, 1)
                    if m
                ],
                dtype=np.uint16,
            )
        else:
            mothers_daughters = np.array([])
            self._log("No mother-daughters assigned")

        return mothers_daughters

    @staticmethod
    def mother_assign_to_mb_matrix(ma: t.List[np.array]):
        """
        Convert from a list of lists of mother-bud paired assignments to a
        sparse matrix with a boolean dtype. The rows correspond to
        to daughter buds. The values are boolean and indicate whether a
        given cell is a mother cell and a given daughter bud is assigned
        to the mother cell in the next timepoint.

        Parameters:
        -----------
        ma : list of lists of integers
            A list of lists of mother-bud assignments. The i-th sublist contains the
            bud assignments for the i-th tile. The integers in each sublist
            represent the mother label, if it is zero no mother was found.

        Returns:
        --------
        mb_matrix : boolean numpy array of shape (n, m)
            An n x m boolean numpy array where n is the total number of cells (sum
            of the lengths of all sublists in ma) and m is the maximum number of buds
            assigned to any mother cell in ma. The value at (i, j) is True if cell i
            is a daughter cell and cell j is its mother assigned to i.

        Examples:
        --------
        ma = [[0, 0, 1], [0, 1, 0]]
        Cells(None).mother_assign_to_mb_matrix(ma)
        # array([[False, False, False, False, False, False],
        #        [False, False, False, False, False, False],
        #        [ True, False, False, False, False, False],
        #        [False, False, False, False, False, False],
        #        [False, False, False,  True, False, False],
        #        [False, False, False, False, False, False]])

        """

        ncells = sum([len(t) for t in ma])
        mb_matrix = np.zeros((ncells, ncells), dtype=bool)
        c = 0
        for cells in ma:
            for d, m in enumerate(cells):
                if m:
                    mb_matrix[c + d, c + m - 1] = True

            c += len(cells)

        return mb_matrix

    @staticmethod
    def mother_assign_from_dynamic(
        ma: np.ndarray, cell_label: t.List[int], trap: t.List[int], ntraps: int
    ) -> t.List[t.List[int]]:
        """
        Interpolate the associated mothers from the 'mother_assign_dynamic' feature.

        Parameters
        ----------
        ma: np.ndarray
            An array with shape (n_t, n_c) containing the 'mother_assign_dynamic' feature.
        cell_label: List[int]
            A list containing the cell labels.
        trap: List[int]
            A list containing the trap labels.
        ntraps: int
            The total number of traps.

        Returns
        -------
        List[List[int]]
            A list of lists containing the interpolated mother assignment for each cell in each trap.
        """
        idlist = list(zip(trap, cell_label))
        cell_gid = np.unique(idlist, axis=0)

        last_lin_preds = [
            find_1st(
                ((cell_label[::-1] == lbl) & (trap[::-1] == tr)),
                True,
                cmp_equal,
            )
            for tr, lbl in cell_gid
        ]
        mother_assign_sorted = ma[::-1][last_lin_preds]

        traps = cell_gid[:, 0]
        iterator = groupby(zip(traps, mother_assign_sorted), lambda x: x[0])
        d = {key: [x[1] for x in group] for key, group in iterator}
        nested_massign = [d.get(i, []) for i in range(ntraps)]

        return nested_massign

    @lru_cache(maxsize=200)
    def labelled_in_frame(
        self, frame: int, global_id: bool = False
    ) -> np.ndarray:
        """
        Returns labels in a 4D ndarray with the global ids with shape
        (ntraps, max_nlabels, ysize, xsize) at a given frame.

        Parameters
        ----------
        frame : int
            The frame number.
        global_id : bool, optional
            If True, the returned array contains global ids, otherwise it
            contains only the local ids of the labels. Default is False.

        Returns
        -------
        np.ndarray
            A 4D numpy array containing the labels in the given frame.
            The array has dimensions (ntraps, max_nlabels, ysize, xsize),
            where max_nlabels is specific for this frame, not the entire
            experiment.

        Notes
        -----
        This method uses lru_cache to cache the results for faster access.

        """
        labels_in_frame = self.labels_at_time(frame)
        n_labels = [
            len(labels_in_frame.get(trap_id, []))
            for trap_id in range(self.ntraps)
        ]
        # maxes = self.max_labels_in_frame(frame)
        stacks_in_frame = self.get_stacks_in_frame(frame, self.tile_size)
        first_id = np.cumsum([0, *n_labels])
        labels_mat = np.zeros(
            (
                self.ntraps,
                max(n_labels),
                *self.tile_size,
            ),
            dtype=int,
        )
        for trap_id, masks in enumerate(stacks_in_frame):  # new_axis = np.pad(
            if trap_id in labels_in_frame:
                new_axis = np.array(labels_in_frame[trap_id], dtype=int)[
                    :, np.newaxis, np.newaxis
                ]
                global_id_masks = new_axis * masks
                if global_id:
                    global_id_masks += first_id[trap_id] * masks
                global_id_masks = np.pad(
                    global_id_masks,
                    pad_width=(
                        (0, labels_mat.shape[1] - global_id_masks.shape[0]),
                        (0, 0),
                        (0, 0),
                    ),
                )
                labels_mat[trap_id] += global_id_masks
        return labels_mat

    def get_stacks_in_frame(
        self, frame: int, tile_shape: t.Tuple[int]
    ) -> t.List[np.ndarray]:
        """
        Returns a list of stacked masks, each corresponding to a tile at a given timepoint.

        Parameters
        ----------
        frame : int
            Frame for which to obtain the stacked masks.
        tile_shape : Tuple[int]
            Shape of a tile to stack the masks into.

        Returns
        -------
        List[np.ndarray]
            List of stacked masks for each tile at the given timepoint.
        """
        masks = self.at_time(frame)
        return [
            stack_masks_in_tile(
                masks.get(trap_id, np.array([], dtype=bool)), tile_shape
            )
            for trap_id in range(self.ntraps)
        ]

    def _sample_tiles_tps(
        self,
        size=1,
        min_consecutive_ntps: int = 15,
        seed: int = 0,
        interval=None,
    ) -> t.Tuple[np.ndarray, np.ndarray]:
        """
        Sample tiles that have a minimum number of cells and are occupied for at least a minimum number of consecutive timepoints.

        Parameters
        ----------
        size: int, optional (default=1)
            The number of tiles to sample.
        min_ncells: int, optional (default=2)
            The minimum number of cells per tile.
        min_consecutive_ntps: int, optional (default=5)
            The minimum number of consecutive timepoints a cell must be present in a trap.
        seed: int, optional (default=0)
            Random seed value for reproducibility.
        interval: None or Tuple(int,int), optional (default=None)
            Random seed value for reproducibility.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray,np.ndarray]
            A tuple of 1D numpy arrays containing the indices of the sampled tiles and the corresponding timepoints.
        """
        # cell_availability_matrix = self.matrix_trap_tp_where(
        #     min_ncells=min_ncells, min_consecutive_tps=min_consecutive_ntps
        # )

        # # Find all valid tiles with min_ncells for at least min_tps
        # tile_ids, tps = np.where(cell_availability_matrix)
        cell_availability_matrix = self.cell_tp_where(
            min_consecutive_tps=min_consecutive_ntps,
            interval=interval,
        )

        # Find all valid tiles with min_ncells for at least min_tps
        index_id, tps = np.where(cell_availability_matrix)

        if interval is None:  # Limit search
            interval = (0, cell_availability_matrix.shape[1])

        np.random.seed(seed)
        choices = np.random.randint(len(index_id), size=size)

        linear_indices = np.zeros_like(self["cell_label"], dtype=bool)
        for cell_index_flat, tp in zip(index_id[choices], tps[choices]):
            tile_id, cell_label = self._flat_index_to_tuple_location(
                cell_index_flat
            )
            linear_indices[
                (
                    (self["cell_label"] == cell_label)
                    & (self["trap"] == tile_id)
                    & (self["timepoint"] == tp)
                )
            ] = True

        return linear_indices

    def _sample_masks(
        self,
        size: int = 1,
        min_consecutive_ntps: int = 15,
        interval: t.Union[None, t.Tuple[int, int]] = None,
        seed: int = 0,
    ) -> t.Tuple[t.Tuple[t.List[int], t.List[int], t.List[int]], np.ndarray]:
        """
        Sample a number of cells from within an interval.

        Parameters
        ----------
        size: int, optional (default=1)
            The number of cells to sample.
        min_ncells: int, optional (default=2)
            The minimum number of cells per tile.
        min_consecutive_ntps: int, optional (default=5)
            The minimum number of consecutive timepoints a cell must be present in a trap.
        seed: int, optional (default=0)
            Random seed value for reproducibility.

        Returns
        -------
        Tuple[Tuple[np.ndarray, np.ndarray, List[int]], List[np.ndarray]]
            Two tuples are returned. The first tuple contains:
            - `tile_ids`: A 1D numpy array of the tile ids that correspond to the tile identifier.
            - `tps`: A 1D numpy array of the timepoints at which the cells were sampled.
            - `cell_ids`: A list of integers that correspond to the local id of the sampled cells.
            The second tuple contains:
            - `masks`: A list of 2D numpy arrays representing the binary masks of the sampled cells at each timepoint.
        """
        sampled_bitmask = self._sample_tiles_tps(
            size=size,
            min_consecutive_ntps=min_consecutive_ntps,
            seed=seed,
            interval=interval,
        )

        #  Sort sampled tiles to use automatic cache when possible
        tile_ids = self["trap"][sampled_bitmask]
        cell_labels = self["cell_label"][sampled_bitmask]
        tps = self["timepoint"][sampled_bitmask]

        masks = []
        for tile_id, cell_label, tp in zip(tile_ids, cell_labels, tps):
            local_idx = self.labels_at_time(tp)[tile_id].index(cell_label)
            tile_mask = self.at_time(tp)[tile_id][local_idx]
            masks.append(tile_mask)

        return (tile_ids, cell_labels, tps), np.stack(masks)

    def matrix_trap_tp_where(
        self, min_ncells: int = 2, min_consecutive_tps: int = 5
    ):
        """
        NOTE CURRENLTY UNUSED WITHIN ALIBY THE MOMENT. MAY BE USEFUL IN THE FUTURE.

        Return a matrix of shape (ntraps x ntps - min_consecutive_tps) to
        indicate traps and time-points where min_ncells are available for at least min_consecutive_tps

        Parameters
        ---------
            min_ncells: int Minimum number of cells
            min_consecutive_tps: int
                Minimum number of time-points a

        Returns
        ---------
            (ntraps x ( ntps-min_consecutive_tps )) 2D boolean numpy array where rows are trap ids and columns are timepoint windows.
            If the value in a cell is true its corresponding trap and timepoint contains more than min_ncells for at least min_consecutive time-points.
        """

        window = sliding_window_view(
            self._tiles_vs_cells_vs_tps, min_consecutive_tps, axis=2
        )
        tp_min = window.sum(axis=-1) == min_consecutive_tps
        ncells_tp_min = tp_min.sum(axis=1) >= min_ncells
        return ncells_tp_min


def stack_masks_in_tile(
    masks: t.List[np.ndarray], tile_shape: t.Tuple[int]
) -> np.ndarray:
    # Stack all masks in a trap padding accordingly if no outlines found
    result = np.zeros((0, *tile_shape), dtype=bool)
    if len(masks):
        result = np.stack(masks)
    return result



#!/usr/bin/env python3

import matplotlib.pyplot as plt

from postprocessor.routines.single_plot import _SinglePlotter


class _SingleBirthPlotter(_SinglePlotter):
    """Draw a line plot of a single time series, but with buddings overlaid"""

    def __init__(
        self,
        trace_timepoints,
        trace_values,
        trace_name,
        birth_mask,
        unit_scaling,
        trace_color,
        birth_color,
        trace_linestyle,
        birth_linestyle,
        xlabel,
        ylabel,
        birth_label,
        plot_title,
    ):
        # Define attributes from arguments
        super().__init__(
            trace_timepoints,
            trace_values,
            trace_name,
            unit_scaling,
            trace_color,
            trace_linestyle,
            xlabel,
            ylabel,
            plot_title,
        )
        # Add some more attributes useful for buddings
        self.birth_mask = birth_mask
        self.birth_color = birth_color
        self.birth_linestyle = birth_linestyle
        self.birth_label = birth_label

    def plot(self, ax):
        """Draw the line plots on the provided Axes."""
        trace_time = self.trace_timepoints * self.unit_scaling
        super().plot(ax)
        birth_mask_bool = self.birth_mask.astype(bool)
        for occurence, birth_time in enumerate(trace_time[birth_mask_bool]):
            if occurence == 0:
                label = self.birth_label
            else:
                label = None
            ax.axvline(
                birth_time,
                color=self.birth_color,
                linestyle=self.birth_linestyle,
                label=label,
            )
        ax.legend()


def single_birth_plot(
    trace_timepoints,
    trace_values,
    trace_name="flavin",
    birth_mask=None,
    unit_scaling=1,
    trace_color="b",
    birth_color="k",
    trace_linestyle="-",
    birth_linestyle="--",
    xlabel="Time (min)",
    ylabel="Normalised flavin fluorescence (AU)",
    birth_label="budding event",
    plot_title="",
    ax=None,
):
    """Plot time series of trace, overlaid with buddings

    Parameters
    ----------
    trace_timepoints : array_like
        Time points (as opposed to the actual times in time units)
    trace_values : array_like
        Trace to plot
    trace_name : string
        Name of trace being plotted, e.g. 'flavin'.
    birth_mask : array_like
        Mask to indicate where buddings are. Expect values of '0' and '1' or
        'False' and 'True' in the elements.
    unit_scaling : int or float
        Unit scaling factor, e.g. 1/60 to convert minutes to hours.
    trace_color : string
        matplotlib colour string for the trace
    birth_color : string
        matplotlib colour string for the vertical lines indicating buddings
    trace_linestyle : string
        matplotlib linestyle argument for the trace
    birth_linestyle : string
        matplotlib linestyle argument for the vertical lines indicating buddings
    xlabel : string
        x axis label.
    ylabel : string
        y axis label.
    birth_label : string
        label for budding event, 'budding event' by default.
    plot_title : string
        Plot title.
    ax : matplotlib Axes
        Axes in which to draw the plot, otherwise use the currently active Axes.

    Returns
    -------
    ax : matplotlib Axes
        Axes object with the plot.

    Examples
    --------
    FIXME: Add docs.

    """
    plotter = _SingleBirthPlotter(
        trace_timepoints,
        trace_values,
        trace_name,
        birth_mask,
        unit_scaling,
        trace_color,
        birth_color,
        trace_linestyle,
        birth_linestyle,
        xlabel,
        ylabel,
        birth_label,
        plot_title,
    )
    if ax is None:
        ax = plt.gca()
    plotter.plot(ax)
    return ax



#!/usr/bin/env jupyter
"""
Visualisation tools useful to generate figures cell pictures and figures from scripts.

These do not depend on matplotlib to work, they focus on array processing.
To check plot-related functions look at plots.py in this folder.
"""
import typing as t
from copy import copy

import numpy as np

from agora.io.cells import Cells
from aliby.io.image import instatiate_image
from aliby.tile.tiler import Tiler, TilerParameters


def fetch_tc(
    image_path: str, results_path: str, t: int = 0, c: int = 0
) -> np.ndarray:
    """
    Return 3D ndarray with (Z,Y,X) for a given pair of time point and channel.
    """
    with instatiate_image(image_path) as img:
        tiler = Tiler.from_h5(img, results_path, TilerParameters.default())
        tc = tiler.get_tp_data(t, c)
    return tc


def get_tiles_at_times(
    image_path: str,
    results_path: str,
    timepoints: t.List[int] = [0],
    tile_reduction: t.Union[
        int, t.List[int], str, t.Callable
    ] = lambda x: concatenate_dims(x, 1, -1),
    channel: int = 1,
) -> np.ndarray:
    """Use Image and tiler to get tiled position for specific time points.

    Parameters
    ----------
    image_path : str
        hdf5 index
    timepoints : t.List[int]
        list of timepoints to fetch
    tile_reduction : t.Union[int, t.List[int], str, t.Callable]
        Reduce dimensionality. Generally used to collapse z-stacks into one

    Examples
    --------
    FIXME: Add docs.


    """

    # Get the correct tile in space and time
    with instatiate_image(image_path) as image:
        tiler = Tiler.from_h5(image, results_path, TilerParameters.default())
        tp_channel_stack = [
            _dispatch_tile_reduction(tile_reduction)(
                tiler.get_tp_data(tp, channel)
            )
            for tp in timepoints
        ]
    return tp_channel_stack


def get_cellmasks_at_times(
    results_path: str, timepoints: t.List[int] = [0]
) -> t.List[t.List[np.ndarray]]:
    return Cells(results_path).at_times(timepoints)


def concatenate_dims(ndarray, axis1: int, axis2: int) -> np.ndarray:
    axis2 = len(ndarray.shape) + axis2 if axis2 < 0 else axis2
    return np.concatenate(np.moveaxis(ndarray, axis1, 0), axis=axis2 - 1)


def get_tile_mask_pairs(
    image_path: str,
    results_path: str,
    timepoints: t.List[int] = [0],
    tile_reduction=lambda x: concatenate_dims(x, 1, -1),
) -> t.Tuple[np.ndarray, t.List[t.List[np.ndarray]]]:

    return (
        get_tiles_at_times(
            image_path, results_path, timepoints, tile_reduction
        ),
        get_cellmasks_at_times(results_path, timepoints),
    )


def _dispatch_tile_reduction(how: t.Union[int, str, t.List[int]], axis=1):
    """
    Return an appropriate dimensional reduction based on the input on a specified axis.
    If "how" is a string, it operates in dimension 1  (to match tile dimension standard Tile, Z, Y, X)
    how: int, str or list of int
        if int or list of int those numbers are indexed;
        if str it assumes it is a numpy function such as np max.
        if it is a callable it applies that operation to the array.
        if None it returns the result as-is
    axis: Only used when "how" is string. Determines the dimension to which the
        standard operation is applied.
    """
    # FUTURE use match case when migrating to python 3.10

    if how is None:
        return lambda x: x
    elif isinstance(how, (int, list)):
        return lambda x: x.take(how, axis=axis)
    elif isinstance(how, str):
        return lambda x: getattr(x, how)(axis=axis)
    elif isinstance(how, t.Callable):
        return lambda x: how(x)
    else:
        raise Exception(f"Invalid reduction {how}")


def tile_like(arr1: np.ndarray, arr2: np.ndarray):
    """
    Tile the first two dimensions of arr1 (ND) to match arr2 (2D)
    """

    result = arr1
    ratio = np.divide(arr2.shape, arr1.shape[-2:]).astype(int)
    if reps := max(ratio - 1):
        tile_ = (
            lambda x, n: np.tile(x, n)
            if ratio.argmax()
            else np.tile(x.T, n + 1).T
        )
        result = np.stack([tile_(mask, reps + 1) for mask in arr1])
    return result


def centre_mask(image: np.ndarray, mask: np.ndarray):
    """Roll image to the centre of the image based on a mask of equal size"""

    cell_centroid = (
        np.max(np.where(mask), axis=1) + np.min(np.where(mask), axis=1)
    ) // 2
    tile_centre = np.array(image.shape) // 2
    return np.roll(image, (tile_centre - cell_centroid), (0, 1))


def long_side_vertical(arr):
    result = arr
    if np.subtract(*arr.shape):
        result = arr.T
    return result


def rescale(arr):
    arr_min = arr.min()
    arr_max = arr.max()
    return (arr - arr_min) / (arr_max - arr_min)


def crop_mask(img: np.ndarray, mask: np.ndarray):
    img = copy(img).astype(float)
    img[~mask] = np.nan
    return img


def _sample_n_tiles_masks(
    image_path: str,
    results_path: str,
    n: int,
    seed: int = 0,
    interval=None,
) -> t.Tuple[t.Tuple, t.Tuple[np.ndarray, np.ndarray]]:

    cells = Cells(results_path)
    indices, masks = cells._sample_masks(n, seed=seed, interval=interval)

    # zipped_indices = zip(*indices)
    for index, (background, cropped_fg) in zip(
        zip(*indices),
        _gen_overlay_masks_tiles(
            image_path,
            results_path,
            masks,
            [indices[i] for i in (0, 2)],
        ),
    ):
        yield index, (background, cropped_fg)


def _overlay_mask_tile(
    image_path: str,
    results_path: str,
    mask: np.ndarray,
    index: t.Tuple[int, int, int],
    bg_channel: int = 0,
    fg_channel: int = 1,
    reduce_z: t.Union[None, t.Callable] = np.max,
) -> t.Tuple[np.ndarray, np.ndarray]:
    """
    Return a tuplw with two channels
    """

    tc = np.stack(
        [
            fetch_tc(image_path, results_path, index[1], i)
            for i in (bg_channel, fg_channel)
        ]
    )  # Returns C(tile)ZYX

    tiles = tc[:, index[0]].astype(float)

    reduced_z = (
        reduce_z(tiles, axis=1) if reduce_z else concatenate_dims(tiles, 1, -2)
    )

    repeated_mask = tile_like(mask, reduced_z[0])

    cropped_fg = crop_mask(reduced_z[1], repeated_mask)

    return reduced_z[0], cropped_fg


def _gen_overlay_masks_tiles(
    image_path: str,
    results_path: str,
    masks: np.ndarray,
    indices: t.Tuple[t.Tuple[int], t.Tuple[int], t.Tuple[int]],
    bg_channel: int = 0,
    fg_channel: int = 1,
    reduce_z: t.Union[None, t.Callable] = np.max,
) -> t.Generator[t.Tuple[np.ndarray, np.ndarray], None, None]:
    """
    Generator of mask tiles
    """

    for mask, index in zip(masks, zip(*indices)):
        yield _overlay_mask_tile(
            image_path,
            results_path,
            mask,
            index,
            bg_channel,
            fg_channel,
            reduce_z,
        )



import bottleneck as bn
import numpy as np
import pandas as pd

from agora.abc import ParametersABC
from postprocessor.core.abc import PostProcessABC


class dsignalParameters(ParametersABC):
    """
    :window: Number of timepoints to consider for signal.
    """

    _defaults = {"window": 10, "min_count": 5}


class dsignal(PostProcessABC):
    """
    Calculate the change in a signal using the mean of a moving window.
    """

    def __init__(self, parameters: dsignalParameters):
        super().__init__(parameters)

    def run(self, signal: pd.DataFrame):
        if signal.shape[1] > self.parameters.window:
            matrix = np.diff(
                bn.move_mean(
                    signal,
                    window=self.parameters.window,
                    min_count=self.parameters.min_count,
                    axis=1,
                ),
                axis=1,
            )
            # Pad values to keep the same signal shape
            matrix = np.pad(matrix, ((0, 0), (0, 1)), constant_values=np.nan)
        else:
            matrix = np.full_like(signal, np.nan)

        return pd.DataFrame(matrix, index=signal.index, columns=signal.columns)



import typing as t

import bottleneck as bn
import numpy as np


def trap_apply(cell_fun, cell_masks, *args, **kwargs):
    """
    Apply a cell_function to a mask and a trap_image.

    Parameters
    ----------
    cell_fun: function
        Function to apply to the cell (from extraction/cell.py)
    cell_masks: 3d array
        Segmentation masks for the cells. Note that cells are in the first dimension (N, Y,X)
    *args: tuple
        Trap_image and any other arguments to pass if needed to custom functions.
    **kwargs: dict
        Keyword arguments to pass if needed to custom functions.
    """
    # apply cell_fun to each cell and return the results as a list
    return [cell_fun(mask, *args, **kwargs) for mask in cell_masks]


def reduce_z(trap_image: np.ndarray, fun: t.Callable, axis: int = 0):
    """
    Reduce the trap_image to 2d.

    Parameters
    ----------
    trap_image: array
        Images for all the channels associated with a trap
    fun: function
        Function to execute the reduction
    axis: int (default 0)
        Axis in which we apply the reduction operation.
    """
    # FUTURE replace with py3.10's match-case.
    if (
        hasattr(fun, "__module__") and fun.__module__[:10] == "bottleneck"
    ):  # Bottleneck type
        return getattr(bn.reduce, fun.__name__)(trap_image, axis=axis)
    elif isinstance(fun, np.ufunc):
        # optimise the reduction function if possible
        return fun.reduce(trap_image, axis=axis)
    else:  # WARNING: Very slow, only use when no alternatives exist
        return np.apply_along_axis(fun, axis, trap_image)



#!/usr/bin/env jupyter
# TODO should this be merged to the regular logfile_parser structure?
"""
Description of new logfile:

All three conditions are concatenated in a single file, in this order:
 - Experiment basic information  (URL in acquisition PC, project, user input)
 - Acquisition settings
 - Experiment start

The section separators are:
-----Acquisition settings-----
-----Experiment started-----

And for a successfully finished experiment we get:

YYYY-MM-DD HH:mm:ss,ms*3 Image acquisition complete WeekDay Mon Day  HH:mm:ss,ms*3 YYYY

For example:
2022-09-30 05:40:59,765 Image acquisition complete Fri Sep 30 05:40:59 2022

Data to extract:
* Basic information
 - Experiment details, which may indicate technical issues
 -  GIT commit
 - (Not working as of 2022/10/03, but projects and tags)
* Basic information
 -

New grammar

- Tables are assumed to end with an empty line.
"""

import logging
import typing as t
from pathlib import PosixPath

import pandas as pd
from pyparsing import (
    CharsNotIn,
    Combine,
    Group,
    Keyword,
    LineEnd,
    LineStart,
    Literal,
    OneOrMore,
    ParserElement,
    Word,
    printables,
)

atomic = t.Union[str, int, float, bool]


class HeaderEndNotFound(Exception):
    def __init__(self, message, errors):
        super().__init__(message)

        self.errors = errors


def extract_header(filepath: PosixPath):
    # header_contents = ""
    with open(filepath, "r") as f:
        try:
            header = ""
            for _ in range(MAX_NLINES):
                line = f.readline()
                header += line
                if HEADER_END in line:
                    break
        except HeaderEndNotFound as e:
            print(f"{MAX_NLINES} checked and no header found")
            raise (e)
        return header


def parse_table(
    string: str,
    start_trigger: t.Union[str, Keyword],
) -> pd.DataFrame:
    """Parse csv-like table

    Parameters
    ----------
    string : str
        contents to parse
    start_trigger : t.Union[str, t.Collection]
        string or triggers that indicate section start.

    Returns
    -------
    pd.Dataframe or dict of atomic values (int,str,bool,float)
        DataFrame representing table.

    Examples
    --------
    >>> table = parse_table()

    """

    if isinstance(start_trigger, str):
        start_trigger: Keyword = Keyword(start_trigger)

    EOL = LineEnd().suppress()
    field = OneOrMore(CharsNotIn(":,\n"))
    line = LineStart() + Group(
        OneOrMore(field + Literal(",").suppress()) + field + EOL
    )
    parser = (
        start_trigger
        + EOL
        + Group(OneOrMore(line))
        + EOL  # end_trigger.suppress()
    )
    parser_result = parser.search_string(string)

    assert all(
        [len(row) == len(parser_result[0]) for row in parser_result]
    ), f"Table {start_trigger} has unequal number of columns"

    assert len(parser_result), f"Parsing is empty. {parser}"
    return table_to_df(parser_result.as_list())


def parse_fields(
    string: str, start_trigger, end_trigger=None
) -> t.Union[pd.DataFrame, t.Dict[str, atomic]]:
    """
    Fields are parsed as key: value

    By default the end is an empty newline.

    For example

    group: YST_1510 field: time
    start: 0
    interval: 300
    frames: 180


    """
    EOL = LineEnd().suppress()

    if end_trigger is None:
        end_trigger = EOL
    elif isinstance(end_trigger, str):
        end_trigger = Literal(end_trigger)

    field = OneOrMore(CharsNotIn(":\n"))
    line = (
        LineStart()
        + Group(field + Combine(OneOrMore(Literal(":").suppress() + field)))
        + EOL
    )
    parser = (
        start_trigger + EOL + Group(OneOrMore(line)) + end_trigger.suppress()
    )
    parser_result = parser.search_string(string)
    results = parser_result.as_list()
    assert len(results), "Parsing returned nothing"
    return fields_to_dict_or_table(results)


# Grammar specification
grammar = {
    "general": {
        "start_trigger": Literal("Swain Lab microscope experiment log file"),
        "type": "fields",
        "end_trigger": "-----Acquisition settings-----",
    },
    "image_config": {
        "start_trigger": "Image Configs:",
        "type": "table",
    },
    "device_properties": {
        "start_trigger": "Device properties:",
        "type": "table",
    },
    "group": {
        "position": {
            "start_trigger": Group(
                Group(Literal("group:") + Word(printables))
                + Group(Literal("field:") + "position")
            ),
            "type": "table",
        },
        **{
            key: {
                "start_trigger": Group(
                    Group(Literal("group:") + Word(printables))
                    + Group(Literal("field:") + key)
                ),
                "type": "fields",
            }
            for key in ("time", "config")
        },
    },
}


ACQ_START = "-----Acquisition settings-----"
HEADER_END = "-----Experiment started-----"
MAX_NLINES = 2000  # In case of malformed logfile
# test_file = "/home/alan/Downloads/pH_med_to_low.log"
# test_file = "/home/alan/Documents/dev/skeletons/scripts/dev/C1_60x.log"

ParserElement.setDefaultWhitespaceChars(" \t")


# time_fields = parse_field(acq, start_trigger=grammar["group"]["time"]["start_trigger"])
# config_fields = parse_fields(
#     acq, start_trigger=grammar["group"]["config"]["start_trigger"]
# )

# general_fields = parse_fields(basic, start_trigger=grammar["general"]["start_trigger"])


def parse_from_grammar(filepath: str, grammar: t.Dict):
    header = extract_header(filepath)
    d = {}
    for key, values in grammar.items():
        try:
            if "type" in values:
                d[key] = parse_x(header, **values)
            else:  # Use subkeys to parse groups
                for subkey, subvalues in values.items():
                    subkey = "_".join((key, subkey))
                    d[subkey] = parse_x(header, **subvalues)
        except Exception as e:
            logging.getLogger("aliby").critical(
                f"Parsing failed for key {key} and values {values}"
            )
            raise (e)
    return d


def table_to_df(result: t.List[t.List]):
    if len(result) > 1:  # Multiple tables with ids to append
        # Generate multiindex from "Name column"
        # index = [row[1][0][1] for table in result for row in table]
        # table[1][0].index("Name") # for automatic indexing
        from itertools import product

        group_name = [
            product((table[0][0][1],), (row[0] for row in table[1][1:]))
            for table in result
        ]
        tmp = [pair for pairset in group_name for pair in pairset]
        multiindices = pd.MultiIndex.from_tuples(tmp)
        df = pd.DataFrame(
            [row for pr in result for row in pr[1][1:]],
            columns=result[0][1][0],
            index=multiindices,
        )
        df.name = result[0][0][1][1]
    else:  # If it is a single table
        df = pd.DataFrame(result[0][1][1:], columns=result[0][1][0])

    return df


def fields_to_dict_or_table(result: t.List[t.List]):
    if len(result) > 1:
        formatted = pd.DataFrame(
            [[row[1] for row in pr[1]] for pr in result],
            columns=[x[0] for x in result[0][1]],
            index=[x[0][0][1] for x in result],
        )

        formatted.name = result[0][0][1][1]

    else:  # If it is a single table
        formatted = {k: _cast_type(v) for k, v in dict(result[0][1]).items()}

    return formatted


def _cast_type(x: str) -> t.Union[str, int, float, bool]:
    # Convert to any possible when possible
    x = x.strip()
    if x.isdigit():
        x = int(x)
    else:
        try:
            x = float(x)
        except:
            try:
                x = ("false", "true").index(x.lower())
            except:
                pass
    return x


def parse_x(string: str, type: str, **kwargs):
    # return eval(f"parse_{type}({string}, **{kwargs})")
    return eval(f"parse_{type}(string, **kwargs)")


def parse_from_swainlab_grammar(filepath: t.Union[str, PosixPath]):
    return parse_from_grammar(filepath, grammar)



"""Set up and run pipelines: tiling, segmentation, extraction, and then post-processing."""
import logging
import os
import re
import traceback
import typing as t
from copy import copy
from importlib.metadata import version
from pathlib import Path, PosixPath

import h5py
import numpy as np
import pandas as pd
from pathos.multiprocessing import Pool
from tqdm import tqdm

from agora.abc import ParametersABC, ProcessABC
from agora.io.metadata import MetaData, parse_logfiles
from agora.io.reader import StateReader
from agora.io.signal import Signal
from agora.io.writer import (
    LinearBabyWriter,
    StateWriter,
    TilerWriter,
)
from aliby.baby_client import BabyParameters, BabyRunner
from aliby.haystack import initialise_tf
from aliby.io.dataset import dispatch_dataset
from aliby.io.image import dispatch_image
from aliby.tile.tiler import Tiler, TilerParameters
from extraction.core.extractor import Extractor, ExtractorParameters
from extraction.core.functions.defaults import exparams_from_meta
from postprocessor.core.processor import PostProcessor, PostProcessorParameters


class PipelineParameters(ParametersABC):
    """Define parameters for the steps of the pipeline."""

    _pool_index = None

    def __init__(
        self, general, tiler, baby, extraction, postprocessing, reporting
    ):
        """Initialise, but called by a class method - not directly."""
        self.general = general
        self.tiler = tiler
        self.baby = baby
        self.extraction = extraction
        self.postprocessing = postprocessing
        self.reporting = reporting

    @classmethod
    def default(
        cls,
        general={},
        tiler={},
        baby={},
        extraction={},
        postprocessing={},
    ):
        """
        Initialise parameters for steps of the pipeline.

        Some parameters are extracted from the log files.

        Parameters
        ---------
        general: dict
            Parameters to set up the pipeline.
        tiler: dict
            Parameters for tiler.
        baby: dict (optional)
            Parameters for Baby.
        extraction: dict (optional)
            Parameters for extraction.
        postprocessing: dict (optional)
            Parameters for post-processing.
        """
        expt_id = general.get("expt_id", 19993)
        if isinstance(expt_id, PosixPath):
            assert expt_id.exists()

            expt_id = str(expt_id)
            general["expt_id"] = expt_id

        directory = Path(general["directory"])

        # get log files, either locally or via OMERO
        with dispatch_dataset(
            expt_id,
            **{k: general.get(k) for k in ("host", "username", "password")},
        ) as conn:
            directory = directory / conn.unique_name
            if not directory.exists():
                directory.mkdir(parents=True)
            # download logs for metadata
            conn.cache_logs(directory)
        try:
            meta_d = MetaData(directory, None).load_logs()
        except Exception as e:
            logging.getLogger("aliby").warn(
                f"WARNING:Metadata: error when loading: {e}"
            )
            minimal_default_meta = {
                "channels": ["Brightfield"],
                "ntps": [2000],
            }
            # set minimal metadata
            meta_d = minimal_default_meta

        # define default values for general parameters
        tps = meta_d.get("ntps", 2000)
        defaults = {
            "general": dict(
                id=expt_id,
                distributed=0,
                tps=tps,
                directory=str(directory.parent),
                filter="",
                earlystop=dict(
                    min_tp=100,
                    thresh_pos_clogged=0.4,
                    thresh_trap_ncells=8,
                    thresh_trap_area=0.9,
                    ntps_to_eval=5,
                ),
                logfile_level="INFO",
                use_explog=True,
            )
        }

        # update default values using inputs
        for k, v in general.items():
            if k not in defaults["general"]:
                defaults["general"][k] = v
            elif isinstance(v, dict):
                for k2, v2 in v.items():
                    defaults["general"][k][k2] = v2
            else:
                defaults["general"][k] = v

        # define defaults and update with any inputs
        defaults["tiler"] = TilerParameters.default(**tiler).to_dict()

        # Generate a backup channel, for when logfile meta is available
        # but not image metadata.
        backup_ref_channel = None
        if "channels" in meta_d and isinstance(
            defaults["tiler"]["ref_channel"], str
        ):
            backup_ref_channel = meta_d["channels"].index(
                defaults["tiler"]["ref_channel"]
            )
        defaults["tiler"]["backup_ref_channel"] = backup_ref_channel

        defaults["baby"] = BabyParameters.default(**baby).to_dict()
        defaults["extraction"] = (
            exparams_from_meta(meta_d)
            or BabyParameters.default(**extraction).to_dict()
        )
        defaults["postprocessing"] = PostProcessorParameters.default(
            **postprocessing
        ).to_dict()
        defaults["reporting"] = {}

        return cls(**{k: v for k, v in defaults.items()})

    def load_logs(self):
        parsed_flattened = parse_logfiles(self.log_dir)
        return parsed_flattened


class Pipeline(ProcessABC):
    """
    Initialise and run tiling, segmentation, extraction and post-processing.

    Each step feeds the next one.

    To customise parameters for any step use the PipelineParameters class.stem
    """

    pipeline_steps = ["tiler", "baby", "extraction"]
    step_sequence = [
        "tiler",
        "baby",
        "extraction",
        "postprocessing",
    ]

    # Specify the group in the h5 files written by each step
    writer_groups = {
        "tiler": ["trap_info"],
        "baby": ["cell_info"],
        "extraction": ["extraction"],
        "postprocessing": ["postprocessing", "modifiers"],
    }
    writers = {  # TODO integrate Extractor and PostProcessing in here
        "tiler": [("tiler", TilerWriter)],
        "baby": [("baby", LinearBabyWriter), ("state", StateWriter)],
    }

    def __init__(self, parameters: PipelineParameters, store=None):
        """Initialise - not usually called directly."""
        super().__init__(parameters)
        if store is not None:
            store = Path(store)
        self.store = store

    @staticmethod
    def setLogger(
        folder, file_level: str = "INFO", stream_level: str = "WARNING"
    ):
        """Initialise and format logger."""
        logger = logging.getLogger("aliby")
        logger.setLevel(getattr(logging, file_level))
        formatter = logging.Formatter(
            "%(asctime)s - %(levelname)s:%(message)s",
            datefmt="%Y-%m-%dT%H:%M:%S%z",
        )
        # for streams - stdout, files, etc.
        ch = logging.StreamHandler()
        ch.setLevel(getattr(logging, stream_level))
        ch.setFormatter(formatter)
        logger.addHandler(ch)
        # create file handler that logs even debug messages
        fh = logging.FileHandler(Path(folder) / "aliby.log", "w+")
        fh.setLevel(getattr(logging, file_level))
        fh.setFormatter(formatter)
        logger.addHandler(fh)

    @classmethod
    def from_yaml(cls, fpath):
        # This is just a convenience function, think before implementing
        # for other processes
        return cls(parameters=PipelineParameters.from_yaml(fpath))

    @classmethod
    def from_folder(cls, dir_path):
        """
        Re-process all h5 files in a given folder.

        All files must share the same parameters, even if they have different channels.

        Parameters
        ---------
        dir_path : str or Pathlib
            Folder containing the files.
        """
        # find h5 files
        dir_path = Path(dir_path)
        files = list(dir_path.rglob("*.h5"))
        assert len(files), "No valid files found in folder"
        fpath = files[0]
        # TODO add support for non-standard unique folder names
        with h5py.File(fpath, "r") as f:
            pipeline_parameters = PipelineParameters.from_yaml(
                f.attrs["parameters"]
            )
        pipeline_parameters.general["directory"] = dir_path.parent
        pipeline_parameters.general["filter"] = [fpath.stem for fpath in files]
        # fix legacy post-processing parameters
        post_process_params = pipeline_parameters.postprocessing.get(
            "parameters", None
        )
        if post_process_params:
            pipeline_parameters.postprocessing["param_sets"] = copy(
                post_process_params
            )
            del pipeline_parameters.postprocessing["parameters"]
        return cls(pipeline_parameters)

    @classmethod
    def from_existing_h5(cls, fpath):
        """
        Re-process an existing h5 file.

        Not suitable for more than one file.

        Parameters
        ---------
        fpath: str
            Name of file.
        """
        with h5py.File(fpath, "r") as f:
            pipeline_parameters = PipelineParameters.from_yaml(
                f.attrs["parameters"]
            )
        directory = Path(fpath).parent
        pipeline_parameters.general["directory"] = directory
        pipeline_parameters.general["filter"] = Path(fpath).stem
        post_process_params = pipeline_parameters.postprocessing.get(
            "parameters", None
        )
        if post_process_params:
            pipeline_parameters.postprocessing["param_sets"] = copy(
                post_process_params
            )
            del pipeline_parameters.postprocessing["parameters"]
        return cls(pipeline_parameters, store=directory)

    @property
    def _logger(self):
        return logging.getLogger("aliby")

    def run(self):
        """Run separate pipelines for all positions in an experiment."""
        # general information in config
        config = self.parameters.to_dict()
        expt_id = config["general"]["id"]
        distributed = config["general"]["distributed"]
        pos_filter = config["general"]["filter"]
        root_dir = Path(config["general"]["directory"])
        self.server_info = {
            k: config["general"].get(k)
            for k in ("host", "username", "password")
        }
        dispatcher = dispatch_dataset(expt_id, **self.server_info)
        logging.getLogger("aliby").info(
            f"Fetching data using {dispatcher.__class__.__name__}"
        )
        # get log files, either locally or via OMERO
        with dispatcher as conn:
            image_ids = conn.get_images()
            directory = self.store or root_dir / conn.unique_name
            if not directory.exists():
                directory.mkdir(parents=True)
            # download logs to use for metadata
            conn.cache_logs(directory)
        # update configuration
        self.parameters.general["directory"] = str(directory)
        config["general"]["directory"] = directory
        self.setLogger(directory)
        # pick particular images if desired
        if pos_filter is not None:
            if isinstance(pos_filter, list):
                image_ids = {
                    k: v
                    for filt in pos_filter
                    for k, v in self.apply_filter(image_ids, filt).items()
                }
            else:
                image_ids = self.apply_filter(image_ids, pos_filter)
        assert len(image_ids), "No images to segment"
        # create pipelines
        if distributed != 0:
            # multiple cores
            with Pool(distributed) as p:
                results = p.map(
                    lambda x: self.run_one_position(*x),
                    [(k, i) for i, k in enumerate(image_ids.items())],
                )
        else:
            # single core
            results = []
            for k, v in tqdm(image_ids.items()):
                r = self.run_one_position((k, v), 1)
                results.append(r)
        return results

    def apply_filter(self, image_ids: dict, filt: int or str):
        """Select images by picking a particular one or by using a regular expression to parse their file names."""
        if isinstance(filt, str):
            # pick images using a regular expression
            image_ids = {
                k: v for k, v in image_ids.items() if re.search(filt, k)
            }
        elif isinstance(filt, int):
            # pick the filt'th image
            image_ids = {
                k: v for i, (k, v) in enumerate(image_ids.items()) if i == filt
            }
        return image_ids

    def run_one_position(
        self,
        name_image_id: t.Tuple[str, str or PosixPath or int],
        index: t.Optional[int] = None,
    ):
        """Set up and run a pipeline for one position."""
        self._pool_index = index
        name, image_id = name_image_id
        # session and filename are defined by calling setup_pipeline.
        # can they be deleted here?
        session = None
        filename = None
        #
        run_kwargs = {"extraction": {"labels": None, "masks": None}}
        try:
            (
                filename,
                meta,
                config,
                process_from,
                tps,
                steps,
                earlystop,
                session,
                trackers_state,
            ) = self._setup_pipeline(image_id)
            loaded_writers = {
                name: writer(filename)
                for k in self.step_sequence
                if k in self.writers
                for name, writer in self.writers[k]
            }
            writer_ow_kwargs = {
                "state": loaded_writers["state"].datatypes.keys(),
                "baby": ["mother_assign"],
            }

            # START PIPELINE
            frac_clogged_traps = 0.0
            min_process_from = min(process_from.values())

            with dispatch_image(image_id)(
                image_id, **self.server_info
            ) as image:
                # initialise steps
                if "tiler" not in steps:
                    steps["tiler"] = Tiler.from_image(
                        image, TilerParameters.from_dict(config["tiler"])
                    )
                if process_from["baby"] < tps:
                    session = initialise_tf(2)
                    steps["baby"] = BabyRunner.from_tiler(
                        BabyParameters.from_dict(config["baby"]),
                        steps["tiler"],
                    )
                    if trackers_state:
                        steps["baby"].crawler.tracker_states = trackers_state
                # limit extraction parameters using the available channels in tiler
                if process_from["extraction"] < tps:
                    # TODO Move this parameter validation into Extractor
                    av_channels = set((*steps["tiler"].channels, "general"))
                    config["extraction"]["tree"] = {
                        k: v
                        for k, v in config["extraction"]["tree"].items()
                        if k in av_channels
                    }
                    config["extraction"]["sub_bg"] = av_channels.intersection(
                        config["extraction"]["sub_bg"]
                    )
                    av_channels_wsub = av_channels.union(
                        [c + "_bgsub" for c in config["extraction"]["sub_bg"]]
                    )
                    tmp = copy(config["extraction"]["multichannel_ops"])
                    for op, (input_ch, _, _) in tmp.items():
                        if not set(input_ch).issubset(av_channels_wsub):
                            del config["extraction"]["multichannel_ops"][op]
                    exparams = ExtractorParameters.from_dict(
                        config["extraction"]
                    )
                    steps["extraction"] = Extractor.from_tiler(
                        exparams, store=filename, tiler=steps["tiler"]
                    )
                    # set up progress meter
                    pbar = tqdm(
                        range(min_process_from, tps),
                        desc=image.name,
                        initial=min_process_from,
                        total=tps,
                    )
                    for i in pbar:
                        if (
                            frac_clogged_traps
                            < earlystop["thresh_pos_clogged"]
                            or i < earlystop["min_tp"]
                        ):
                            # run through steps
                            for step in self.pipeline_steps:
                                if i >= process_from[step]:
                                    result = steps[step].run_tp(
                                        i, **run_kwargs.get(step, {})
                                    )
                                    if step in loaded_writers:
                                        loaded_writers[step].write(
                                            data=result,
                                            overwrite=writer_ow_kwargs.get(
                                                step, []
                                            ),
                                            tp=i,
                                            meta={"last_processed": i},
                                        )
                                    # perform step
                                    if (
                                        step == "tiler"
                                        and i == min_process_from
                                    ):
                                        logging.getLogger("aliby").info(
                                            f"Found {steps['tiler'].n_tiles} traps in {image.name}"
                                        )
                                    elif step == "baby":
                                        # write state and pass info to Extractor
                                        loaded_writers["state"].write(
                                            data=steps[
                                                step
                                            ].crawler.tracker_states,
                                            overwrite=loaded_writers[
                                                "state"
                                            ].datatypes.keys(),
                                            tp=i,
                                        )
                                    elif step == "extraction":
                                        # remove mask/label after extraction
                                        for k in ["masks", "labels"]:
                                            run_kwargs[step][k] = None
                            # check and report clogging
                            frac_clogged_traps = self.check_earlystop(
                                filename, earlystop, steps["tiler"].tile_size
                            )
                            self._log(
                                f"{name}:Clogged_traps:{frac_clogged_traps}"
                            )
                            frac = np.round(frac_clogged_traps * 100)
                            pbar.set_postfix_str(f"{frac} Clogged")
                        else:
                            # stop if too many traps are clogged
                            self._log(
                                f"{name}:Stopped early at time {i} with {frac_clogged_traps} clogged traps"
                            )
                            meta.add_fields({"end_status": "Clogged"})
                            break
                        meta.add_fields({"last_processed": i})
                    # run post-processing
                    meta.add_fields({"end_status": "Success"})
                    post_proc_params = PostProcessorParameters.from_dict(
                        config["postprocessing"]
                    )
                    PostProcessor(filename, post_proc_params).run()
                    self._log("Analysis finished successfully.", "info")
                    return 1

        except Exception as e:
            # catch bugs during setup or run time
            logging.exception(
                f"{name}: Exception caught.",
                exc_info=True,
            )
            # print the type, value, and stack trace of the exception
            traceback.print_exc()
            raise e
        finally:
            _close_session(session)

    @staticmethod
    def check_earlystop(filename: str, es_parameters: dict, tile_size: int):
        """
        Check recent time points for tiles with too many cells.

        Returns the fraction of clogged tiles, where clogged tiles have
        too many cells or too much of their area covered by cells.

        Parameters
        ----------
        filename: str
            Name of h5 file.
        es_parameters: dict
            Parameters defining when early stopping should happen.
            For example:
                    {'min_tp': 100,
                    'thresh_pos_clogged': 0.4,
                    'thresh_trap_ncells': 8,
                    'thresh_trap_area': 0.9,
                    'ntps_to_eval': 5}
        tile_size: int
            Size of tile.
        """
        # get the area of the cells organised by trap and cell number
        s = Signal(filename)
        df = s["/extraction/general/None/area"]
        # check the latest time points only
        cells_used = df[
            df.columns[-1 - es_parameters["ntps_to_eval"] : -1]
        ].dropna(how="all")
        # find tiles with too many cells
        traps_above_nthresh = (
            cells_used.groupby("trap").count().apply(np.mean, axis=1)
            > es_parameters["thresh_trap_ncells"]
        )
        # find tiles with cells covering too great a fraction of the tiles' area
        traps_above_athresh = (
            cells_used.groupby("trap").sum().apply(np.mean, axis=1)
            / tile_size**2
            > es_parameters["thresh_trap_area"]
        )
        return (traps_above_nthresh & traps_above_athresh).mean()

    # FIXME: Remove this functionality. It used to be for
    # older hdf5 file formats.
    def _load_config_from_file(
        self,
        filename: PosixPath,
        process_from: t.Dict[str, int],
        trackers_state: t.List,
        overwrite: t.Dict[str, bool],
    ):
        with h5py.File(filename, "r") as f:
            for k in process_from.keys():
                if not overwrite[k]:
                    process_from[k] = self.legacy_get_last_tp[k](f)
                    process_from[k] += 1
        return process_from, trackers_state, overwrite

    # FIXME: Remove this functionality. It used to be for
    # older hdf5 file formats.
    @staticmethod
    def legacy_get_last_tp(step: str) -> t.Callable:
        """Get last time-point in different ways depending
        on which step we are using

        To support segmentation in aliby < v0.24
        TODO Deprecate and replace with State method
        """
        switch_case = {
            "tiler": lambda f: f["trap_info/drifts"].shape[0] - 1,
            "baby": lambda f: f["cell_info/timepoint"][-1],
            "extraction": lambda f: f[
                "extraction/general/None/area/timepoint"
            ][-1],
        }
        return switch_case[step]

    def _setup_pipeline(
        self, image_id: int
    ) -> t.Tuple[
        PosixPath,
        MetaData,
        t.Dict,
        int,
        t.Dict,
        t.Dict,
        t.Optional[int],
        t.List[np.ndarray],
    ]:
        """
        Initialise steps in a pipeline.

        If necessary use a file to re-start experiments already partly run.

        Parameters
        ----------
        image_id : int or str
            Identifier of a data set in an OMERO server or a filename.

        Returns
        -------
        filename: str
            Path to a h5 file to write to.
        meta: object
            agora.io.metadata.MetaData object
        config: dict
            Configuration parameters.
        process_from: dict
            Gives from which time point each step of the pipeline should start.
        tps: int
            Number of time points.
        steps: dict
        earlystop: dict
            Parameters to check whether the pipeline should be stopped.
        session: None
        trackers_state: list
            States of any trackers from earlier runs.
        """
        config = self.parameters.to_dict()
        # TODO Alan: Verify if session must be passed
        session = None
        earlystop = config["general"].get("earlystop", None)
        process_from = {k: 0 for k in self.pipeline_steps}
        steps = {}
        # check overwriting
        ow_id = config["general"].get("overwrite", 0)
        ow = {step: True for step in self.step_sequence}
        if ow_id and ow_id is not True:
            ow = {
                step: self.step_sequence.index(ow_id) < i
                for i, step in enumerate(self.step_sequence, 1)
            }

        # Set up
        directory = config["general"]["directory"]

        trackers_state: t.List[np.ndarray] = []
        with dispatch_image(image_id)(image_id, **self.server_info) as image:
            filename = Path(f"{directory}/{image.name}.h5")
            meta = MetaData(directory, filename)
            from_start = True if np.any(ow.values()) else False
            # remove existing file if overwriting
            if (
                from_start
                and (
                    config["general"].get("overwrite", False)
                    or np.all(list(ow.values()))
                )
                and filename.exists()
            ):
                os.remove(filename)
            # if the file exists with no previous segmentation use its tiler
            if filename.exists():
                self._log("Result file exists.", "info")
                if not ow["tiler"]:
                    steps["tiler"] = Tiler.from_hdf5(image, filename)
                    try:
                        (
                            process_from,
                            trackers_state,
                            ow,
                        ) = self._load_config_from_file(
                            filename, process_from, trackers_state, ow
                        )
                        # get state array
                        trackers_state = (
                            []
                            if ow["baby"]
                            else StateReader(filename).get_formatted_states()
                        )
                        config["tiler"] = steps["tiler"].parameters.to_dict()
                    except Exception:
                        self._log(f"Overwriting tiling data")

            if config["general"]["use_explog"]:
                meta.run()
            # add metadata not in the log file
            meta.add_fields(
                {
                    "aliby_version": version("aliby"),
                    "baby_version": version("aliby-baby"),
                    "omero_id": config["general"]["id"],
                    "image_id": image_id
                    if isinstance(image_id, int)
                    else str(image_id),
                    "parameters": PipelineParameters.from_dict(
                        config
                    ).to_yaml(),
                }
            )
            tps = min(config["general"]["tps"], image.data.shape[0])
            return (
                filename,
                meta,
                config,
                process_from,
                tps,
                steps,
                earlystop,
                session,
                trackers_state,
            )


def _close_session(session):
    if session:
        session.close()



#!/usr/bin/env jupyter

"""
Convert some types to others
"""


def _str_to_int(x: str or None):
    """
    Cast string as int if possible. If Nonetype return None.
    """
    if x:
        try:
            return int(x)
        except:
            return x



#!/usr/bin/env python3

import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import seaborn as sns

from postprocessor.routines.plottingabc import BasePlotter


class _BoxplotPlotter(BasePlotter):
    """Draw boxplots over time"""

    def __init__(
        self,
        trace_df,
        trace_name,
        unit_scaling,
        box_color,
        xtick_step,
        xlabel,
        plot_title,
    ):
        super().__init__(trace_name, unit_scaling, xlabel, plot_title)
        # Define attributes from arguments
        self.trace_df = trace_df
        self.box_color = box_color
        self.xtick_step = xtick_step

        # Define some labels
        self.ylabel = "Normalised " + self.trace_name + " fluorescence (AU)"

        # Define horizontal axis ticks and labels
        # hacky! -- redefine column names
        trace_df.columns = trace_df.columns * self.unit_scaling
        self.fmt = ticker.FuncFormatter(
            lambda x, pos: "{0:g}".format(
                x / (self.xtick_step / self.unit_scaling)
            )
        )

    def plot(self, ax):
        """Draw the heatmap on the provided Axes."""
        super().plot(ax)
        ax.xaxis.set_major_formatter(self.fmt)
        sns.boxplot(
            data=self.trace_df,
            color=self.box_color,
            linewidth=1,
            ax=ax,
        )
        ax.xaxis.set_major_locator(
            ticker.MultipleLocator(self.xtick_step / self.unit_scaling)
        )


def boxplot(
    trace_df,
    trace_name,
    unit_scaling=1,
    box_color="b",
    xtick_step=60,
    xlabel="Time (min)",
    plot_title="",
    ax=None,
):
    """Draw series of boxplots from an array of time series of traces

    Draw series of boxplots from an array of time series of traces, showing the
    distribution of values at each time point over time.

    Parameters
    ----------
    trace_df : pandas.DataFrame
        Time series of traces (rows = cells, columns = time points).
    trace_name : string
        Name of trace being plotted, e.g. 'flavin'.
    unit_scaling : int or float
        Unit scaling factor, e.g. 1/60 to convert minutes to hours.
    box_color : string
        matplolib colour string, specifies colour of boxes in boxplot
    xtick_step : int or float
        Interval length, in unit time, to draw x axis ticks.
    xlabel : string
        x axis label.
    plot_title : string
        Plot title.
    ax : matplotlib Axes
        Axes in which to draw the plot, otherwise use the currently active Axes.

    Returns
    -------
    ax : matplotlib Axes
        Axes object with the heatmap.

    Examples
    --------
    FIXME: Add docs.

    """

    plotter = _BoxplotPlotter(
        trace_df,
        trace_name,
        unit_scaling,
        box_color,
        xtick_step,
        xlabel,
        plot_title,
    )
    if ax is None:
        ax = plt.gca()
    plotter.plot(ax)
    return ax



"""Functions for identifying and dealing with ALCATRAS traps."""

import numpy as np
from skimage import feature, transform
from skimage.filters import threshold_otsu
from skimage.filters.rank import entropy
from skimage.measure import label, regionprops
from skimage.morphology import closing, disk, square
from skimage.segmentation import clear_border
from skimage.util import img_as_ubyte


def half_floor(x, tile_size):
    return x - tile_size // 2


def half_ceil(x, tile_size):
    return x + -(tile_size // -2)


def segment_traps(
    image,
    tile_size,
    downscale=0.4,
    disk_radius_frac=0.01,
    square_size=3,
    min_frac_tilesize=0.3,
    **identify_traps_kwargs,
):
    """
    Use an entropy filter and Otsu thresholding to find a trap template,
    which is then passed to identify_trap_locations.

    To obtain candidate traps the major axis length of a tile must be smaller than tilesize.

    The hyperparameters have not been optimised.

    Parameters
    ----------
    image: 2D array
    tile_size: integer
        Size of the tile
    downscale: float (optional)
        Fraction by which to shrink image
    disk_radius_frac: float (optional)
        Radius of disk using in the entropy filter
    square_size: integer (optional)
        Parameter for a morphological closing applied to thresholded
        image
    min_frac_tilesize: float (optional)
    max_frac_tilesize: float (optional)
        Used to determine bounds on the major axis length of regions
        suspected of containing traps.
    identify_traps_kwargs:
        Passed to identify_trap_locations

    Returns
    -------
    traps: an array of pairs of integers
        The coordinates of the centroids of the traps.
    """
    # keep a memory of image in case need to re-run
    img = image
    # bounds on major axis length of traps
    min_mal = min_frac_tilesize * tile_size

    # shrink image
    if downscale != 1:
        img = transform.rescale(image, downscale)
    # generate an entropy image using a disk footprint
    disk_radius = int(min([disk_radius_frac * x for x in img.shape]))
    entropy_image = entropy(img_as_ubyte(img), disk(disk_radius))
    if downscale != 1:
        entropy_image = transform.rescale(entropy_image, 1 / downscale)
    # find Otsu threshold for entropy image
    thresh = threshold_otsu(entropy_image)
    # apply morphological closing to thresholded, and so binary, image
    bw = closing(entropy_image > thresh, square(square_size))
    # remove artifacts connected to image border
    cleared = clear_border(bw)

    # label distinct regions of the image
    label_image = label(cleared)
    # find regions likely to contain traps:
    # with a major axis length within a certain range
    # and a centroid at least tile_size // 2 away from the image edge
    idx_valid_region = [
        (i, region)
        for i, region in enumerate(regionprops(label_image))
        if min_mal < region.major_axis_length < tile_size
        and tile_size // 2
        < region.centroid[0]
        < half_floor(image.shape[0], tile_size) - 1
        and tile_size // 2
        < region.centroid[1]
        < half_floor(image.shape[1], tile_size) - 1
    ]

    assert idx_valid_region, "No valid tiling regions found"

    _, valid_region = zip(*idx_valid_region)

    # find centroids and minor axes lengths of valid regions
    centroids = (
        np.array([x.centroid for x in valid_region]).round().astype(int)
    )
    minals = [region.minor_axis_length for region in valid_region]
    # coords for best trap
    x, y = np.round(centroids[np.argmin(minals)]).astype(int)

    # make candidate templates from the other traps found
    candidate_templates = [
        image[
            half_floor(x, tile_size) : half_ceil(x, tile_size),
            half_floor(y, tile_size) : half_ceil(y, tile_size),
        ]
        for x, y in centroids
    ]
    # make a mean template from all the found traps
    mean_template = np.stack(candidate_templates).astype(int).mean(axis=0)

    # find traps using the mean trap template
    traps = identify_trap_locations(
        image, mean_template, **identify_traps_kwargs
    )

    # if there are too few traps, try again
    traps_retry = []
    if len(traps) < 30 and downscale != 1:
        print("Tiler:TrapIdentification: Trying again.")
        traps_retry = segment_traps(image, tile_size, downscale=1)

    # return results with the most number of traps
    if len(traps_retry) < len(traps):
        return traps
    else:
        return traps_retry


def identify_trap_locations(
    image, trap_template, optimize_scale=True, downscale=0.35, trap_size=None
):
    """
    Identify the traps in a single image based on a trap template.

    Requires the trap template to be similar to the image
    (same camera, same magnification - ideally the same experiment).

    Use normalised correlation in scikit-image's to match_template.

    The search is sped up by down-scaling both the image and
    the trap template before running the template matching.

    The trap template is rotated and re-scaled to improve matching.
    The parameters of the rotation and re-scaling are optimised, although
    over restricted ranges.

    Parameters
    ----------
    image: 2D array
    trap_template: 2D array
    optimize_scale : boolean (optional)
    downscale: float (optional)
        Fraction by which to downscale to increase speed
    trap_size: integer (optional)
        If unspecified, the size is determined from the trap_template

    Returns
    -------
    coordinates: an array of pairs of integers
    """
    if trap_size is None:
        trap_size = trap_template.shape[0]
    # careful: the image is float16!
    img = transform.rescale(image.astype(float), downscale)
    template = transform.rescale(trap_template, downscale)

    # try multiple rotations of template to determine
    # which best matches the image
    # result is squared because the sign of the correlation is unimportant
    matches = {
        rotation: feature.match_template(
            img,
            transform.rotate(template, rotation, cval=np.median(img)),
            pad_input=True,
            mode="median",
        )
        ** 2
        for rotation in [0, 90, 180, 270]
    }
    # find best rotation
    best_rotation = max(matches, key=lambda x: np.percentile(matches[x], 99.9))
    # rotate template by best rotation
    template = transform.rotate(template, best_rotation, cval=np.median(img))

    if optimize_scale:
        # try multiple scales appled to template to determine which
        # best matches the image
        scales = np.linspace(0.5, 2, 10)
        matches = {
            scale: feature.match_template(
                img,
                transform.rescale(template, scale),
                mode="median",
                pad_input=True,
            )
            ** 2
            for scale in scales
        }
        # find best scale
        best_scale = max(
            matches, key=lambda x: np.percentile(matches[x], 99.9)
        )
        # choose the best result - an image of normalised correlations
        # with the template
        matched = matches[best_scale]
    else:
        # find the image of normalised correlations with the template
        matched = feature.match_template(
            img, template, pad_input=True, mode="median"
        )
    # re-scale back the image of normalised correlations
    # find the coordinates of local maxima
    coordinates = feature.peak_local_max(
        transform.rescale(matched, 1 / downscale),
        min_distance=int(trap_size * 0.70),
        exclude_border=(trap_size // 3),
    )
    return coordinates



from typing import Dict, Tuple

import numpy as np
import pandas as pd
from agora.utils.lineage import mb_array_to_dict

from postprocessor.core.lineageprocess import (
    LineageProcess,
    LineageProcessParameters,
)


class BudMetricParameters(LineageProcessParameters):
    """
    Parameters
    """

    _defaults = {"lineage_location": "postprocessing/lineage_merged"}


class BudMetric(LineageProcess):
    """
    Requires mother-bud information to create a new dataframe where the indices are mother ids and
    values are the daughters' values for a given signal.
    """

    def __init__(self, parameters: BudMetricParameters):
        super().__init__(parameters)

    def run(
        self,
        signal: pd.DataFrame,
        mother_bud_ids: Dict[pd.Index, Tuple[pd.Index]] = None,
    ):
        if mother_bud_ids is None:
            mother_bud_ids = mb_array_to_dict(self.lineage)

        return self.get_bud_metric(signal, mother_bud_ids)

    @staticmethod
    def get_bud_metric(signal: pd.DataFrame, md: Dict[Tuple, Tuple[Tuple]]):
        """

        signal: Daughter-inclusive dataframe
        md: Mother-daughters dictionary where key is mother's index and value a list of daugher indices

        Get fvi (First Valid Index) for all cells
        Create empty matrix
        for every mother:
         - Get daughters' subdataframe
         - sort  daughters by cell label
         - get series of fvis
         - concatenate the values of these ranges from the dataframe
        Fill the empty matrix
        Convert matrix into dataframe using mother indices

        """
        mothers_mat = np.zeros((len(md), signal.shape[1]))
        for i, daughters in enumerate(md.values()):
            dau_vals = signal.loc[set(daughters)].droplevel("trap")
            sorted_da_ids = dau_vals.sort_index(level="cell_label")
            tp_fvt = sorted_da_ids.apply(
                lambda x: x.last_valid_index(), axis=0
            )

            tp_fvt = sorted_da_ids.index.get_indexer(tp_fvt)
            tp_fvt[tp_fvt < 0] = sorted_da_ids.shape[0] - 1

            buds_metric = np.choose(tp_fvt, sorted_da_ids.values)
            # mothers_mat[i, tp_fvt[0] : tp_fvt[0] + len(buds_metric)] = buds_metric
            mothers_mat[i] = buds_metric

        df = pd.DataFrame(mothers_mat, index=md.keys(), columns=signal.columns)
        if len(df):
            df.index.names = signal.index.names
        return df



#!/usr/bin/env jupyter
from importlib_resources import files
from logfile_parser import Parser

grammars_dir = files("logfile_parser") / "grammars"


def get_examples_dir():
    return files("aliby").parent.parent / "examples" / "logfile_parser"


def get_logfile_grammars_dir():
    return files("logfile_parser") / "grammars"


def get_legacy_log_example_interface() -> dict:
    parsed = {}
    examples_dir = get_examples_dir()
    grammars_dir = get_logfile_grammars_dir()

    for gram in ("acq", "log"):
        for gram_file in grammars_dir.glob(f"multiDGUI_{gram}_format.json"):
            parser = Parser(gram_file)
            for file_to_parse in examples_dir.glob(f"*{gram}.txt"):
                with open(file_to_parse, "r") as f:
                    parsed = {**parsed, **parser.parse(f)}
    return parsed


def to_legacy(parsed_logfile: dict) -> dict:
    """
    Convert the output of the new logfile parsing to legacy to a minimal working set of metadata.
    This converts the new more complex metadata structure to the previous one that did not have configuration profiles, but instead one configuration per channel.
    This is a temporal solution as we transition into a more general metadata structure that accounts for heterogeneous groups.

    We convert image configs to channels, and add general metadata to the root.
    """
    name_translation = {
        "Microscope name": "microscope",
    }
    channel_name_translation = {
        "Image config": "channel",
        "Channel": "channel_hardware",
        "Exposure (ms)": "exposure",
        "Z spacing (um)": "zsect",
    }

    # Translate general data
    general = {v: d["general"][k] for k, v in name_translation.items()}
    # Translate and cast image configs
    channels = {
        v: list(map(_cast_type, parsed_logfile["image_config"][k]))
        for k, v in channel_name_translation.items()
    }
    legacy_format = {"channels": channels, **general}
    return legacy_format



#!/usr/bin/env python3

import matplotlib.pyplot as plt
import numpy as np


class _HistogramPlotter:
    """Draw histogram"""

    def __init__(
        self,
        values,
        label,
        color,
        binsize,
        lognormal,
        lognormal_base,
        xlabel,
        ylabel,
        plot_title,
    ):
        # Define attributes from arguments
        self.values = values
        self.label = label
        self.color = color
        self.binsize = binsize
        self.lognormal = lognormal
        self.lognormal_base = lognormal_base
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.plot_title = plot_title

        # Define median
        self.median = np.median(self.values)

        # Define bins
        if self.lognormal:
            self.bins = np.logspace(
                0,
                np.ceil(
                    np.log(np.nanmax(values)) / np.log(self.lognormal_base)
                ),
                base=self.lognormal_base,
            )  # number of bins will be 50 by default, as it's the default in np.logspace
        else:
            self.bins = np.arange(
                self.binsize * (np.nanmin(values) // self.binsize - 2),
                self.binsize * (np.nanmax(values) // self.binsize + 2),
                self.binsize,
            )

    def plot(self, ax):
        """Plot histogram onto specified Axes."""
        ax.set_ylabel(self.ylabel)
        ax.set_xlabel(self.xlabel)
        ax.set_title(self.plot_title)

        if self.lognormal:
            ax.set_xscale("log")
        ax.hist(
            self.values,
            self.bins,
            alpha=0.5,
            color=self.color,
            label=self.label,
        )
        ax.axvline(
            self.median,
            color=self.color,
            alpha=0.75,
            label="median " + self.label,
        )
        ax.legend(loc="upper right")


def histogram(
    values,
    label,
    color="b",
    binsize=5,
    lognormal=False,
    lognormal_base=10,
    xlabel="Time (min)",
    ylabel="Number of occurences",
    plot_title="Distribution",
    ax=None,
):
    """Draw histogram with median indicated

    Parameters
    ----------
    values : array_like
        Input values for histogram
    label : string
        Name of value being plotting, e.g. cell division cycle length.
    color : string
        Colour of bars.
    binsize : float
        Bin size.
    lognormal : bool
        Whether to use a log scale for the horizontal axis.
    lognormal_base : float
        Base of the log scale, if lognormal is True.
    xlabel : string
        x axis label.
    ylabel : string
        y axis label.
    plot_title : string
        Plot title.
    ax : matplotlib Axes
        Axes in which to draw the plot, otherwise use the currently active Axes.

    Returns
    -------
    ax : matplotlib Axes
        Axes object with the histogram.

    Examples
    --------
    FIXME: Add docs.

    """
    plotter = _HistogramPlotter(
        values,
        label,
        color,
        binsize,
        lognormal,
        lognormal_base,
        xlabel,
        ylabel,
        plot_title,
    )
    if ax is None:
        ax = plt.gca()
    plotter.plot(ax)
    return ax



#!/usr/bin/env python3

import typing as t
from itertools import product

import numpy as np
import pandas as pd

from postprocessor.core.lineageprocess import (
    LineageProcess,
    LineageProcessParameters,
)


class buddingsParameters(LineageProcessParameters):
    """Parameter class to obtain budding events.

    Parameters
    ----------
    LineageProcessParameters : lineage_location
        Location of lineage matrix to be used for calculations.

    Examples
    --------
    FIXME: Add docs.

    """

    _defaults = {"lineage_location": "postprocessing/lineage_merged"}


class buddings(LineageProcess):
    """
    Calculate buddings in a trap assuming one mother per trap
    returns a pandas series with the buddings.

    We define a budding event as the moment in which a bud was identified for
    the first time, even if the bud is not considered one until later
    in the experiment.
    """

    def __init__(self, parameters: buddingsParameters):
        super().__init__(parameters)

    def load_lineage(self, lineage):
        self.lineage = lineage

    def run(
        self, signal: pd.DataFrame, lineage: np.ndarray = None
    ) -> pd.DataFrame:
        lineage = lineage or self.lineage

        # Get time of first appearance for all cells
        fvi = signal.apply(lambda x: x.first_valid_index(), axis=1)

        # Select mother cells in a given dataset
        traps_mothers: t.Dict[tuple, list] = {
            tuple(mo): [] for mo in lineage[:, :2] if tuple(mo) in signal.index
        }
        for trap, mother, daughter in lineage:
            if (trap, mother) in traps_mothers.keys():
                traps_mothers[(trap, mother)].append(daughter)

        mothers = signal.loc[
            set(signal.index).intersection(traps_mothers.keys())
        ]
        # Create a new dataframe with dimensions (n_mother_cells * n_timepoints)
        buddings = pd.DataFrame(
            np.zeros((mothers.shape[0], signal.shape[1])).astype(bool),
            index=mothers.index,
            columns=signal.columns,
        )
        buddings.columns.names = ["timepoint"]

        # Fill the budding events
        for mother_id, daughters in traps_mothers.items():
            daughters_idx = set(
                fvi.loc[
                    fvi.index.intersection(
                        list(product((mother_id[0],), daughters))
                    )
                ].values
            ).difference({0})
            buddings.loc[
                mother_id,
                daughters_idx,
            ] = True

        return buddings



#!/usr/bin/env jupyter
import argparse

from agora.utils.cast import _str_to_int
from aliby.pipeline import Pipeline, PipelineParameters


def run():
    """
    Run a default microscopy analysis pipeline.

    Parse command-line arguments and set default parameter values for running a pipeline, then
    construct and execute the pipeline with the parameters obtained. Command-line arguments can
    override default parameter values. If a command-line argument is a string representation of
    an integer, convert it to an integer.

    Returns
    -------
    None

    Examples
    --------
    FIXME: Add docs.
    """
    parser = argparse.ArgumentParser(
        prog="aliby-run",
        description="Run a default microscopy analysis pipeline",
    )

    param_values = {
        "expt_id": None,
        "distributed": 2,
        "tps": 2,
        "directory": "./data",
        "filter": 0,
        "host": None,
        "username": None,
        "password": None,
    }

    for k in param_values:
        parser.add_argument(f"--{k}", action="store")

    args = parser.parse_args()

    for k in param_values:
        if passed_value := _str_to_int(getattr(args, k)):

            param_values[k] = passed_value

    params = PipelineParameters.default(general=param_values)
    p = Pipeline(params)

    p.run()



"""
Anthology of interfaces for different parsers and lack of them.

ALIBY decides on using different metadata parsers based on two elements:

1. The parameter given by PipelineParameters (Either True/False, or a string pointing to the metadata file)
2. The available files in the root folder where images are found (remote or locally)

If parameters is a string pointing to a metadata file, ALIBY picks a parser based on the file format.
If parameters is True (as a boolean), ALIBY searches for any available file and uses the first valid one.
If there are no metadata files, ALIBY requires indicating indices for tiler, segmentation and extraction.

"""
import glob
import logging
import os
import typing as t
from datetime import datetime
from pathlib import Path, PosixPath

import pandas as pd
from pytz import timezone

from agora.io.writer import Writer
from logfile_parser import Parser
from logfile_parser.swainlab_parser import parse_from_swainlab_grammar


class MetaData:
    """Small metadata Process that loads log."""

    def __init__(self, log_dir, store):
        self.log_dir = log_dir
        self.store = store
        self.metadata_writer = Writer(self.store)

    def __getitem__(self, item):
        return self.load_logs()[item]

    def load_logs(self):
        # parsed_flattened = parse_logfiles(self.log_dir)
        parsed_flattened = dispatch_metadata_parser(self.log_dir)
        return parsed_flattened

    def run(self, overwrite=False):
        metadata_dict = self.load_logs()
        self.metadata_writer.write(
            path="/", meta=metadata_dict, overwrite=overwrite
        )

    def add_field(self, field_name, field_value, **kwargs):
        self.metadata_writer.write(
            path="/",
            meta={field_name: field_value},
            **kwargs,
        )

    def add_fields(self, fields_values: dict, **kwargs):
        for field, value in fields_values.items():
            self.add_field(field, value)


# Paradigm: able to do something with all datatypes present in log files,
# then pare down on what specific information is really useful later.

# Needed because HDF5 attributes do not support dictionaries
def flatten_dict(nested_dict, separator="/"):
    """
    Flattens nested dictionary. If empty return as-is.
    """
    flattened = {}
    if nested_dict:
        df = pd.json_normalize(nested_dict, sep=separator)
        flattened = df.to_dict(orient="records")[0] or {}

    return flattened


# Needed because HDF5 attributes do not support datetime objects
# Takes care of time zones & daylight saving
def datetime_to_timestamp(time, locale="Europe/London"):
    """
    Convert datetime object to UNIX timestamp
    """
    return timezone(locale).localize(time).timestamp()


def find_file(root_dir, regex):
    file = [
        f
        for f in glob.glob(os.path.join(str(root_dir), regex))
        if Path(f).name != "aliby.log"  # Skip filename reserved for aliby
    ]

    if len(file) > 1:
        print(
            "Warning:Metadata: More than one logfile found. Defaulting to first option."
        )
        file = [sorted(file)[0]]
    if len(file) == 0:
        logging.getLogger("aliby").log(
            logging.WARNING, "Metadata: No valid swainlab .log found."
        )
    else:
        return file[0]
    return None


# TODO: re-write this as a class if appropriate
# WARNING: grammars depend on the directory structure of a locally installed
# logfile_parser repo
def parse_logfiles(
    root_dir,
    acq_grammar="multiDGUI_acq_format.json",
    log_grammar="multiDGUI_log_format.json",
):
    """
    Parse acq and log files depending on the grammar specified, then merge into
    single dict.
    """
    # Both acq and log files contain useful information.
    # ACQ_FILE = 'flavin_htb2_glucose_long_ramp_DelftAcq.txt'
    # LOG_FILE = 'flavin_htb2_glucose_long_ramp_Delftlog.txt'
    log_parser = Parser(log_grammar)
    acq_parser = Parser(acq_grammar)

    log_file = find_file(root_dir, "*log.txt")
    acq_file = find_file(root_dir, "*[Aa]cq.txt")

    parsed = {}
    if log_file and acq_file:
        with open(log_file, "r") as f:
            log_parsed = log_parser.parse(f)

        with open(acq_file, "r") as f:
            acq_parsed = acq_parser.parse(f)

        parsed = {**acq_parsed, **log_parsed}

    for key, value in parsed.items():
        if isinstance(value, datetime):
            parsed[key] = datetime_to_timestamp(value)

    parsed_flattened = flatten_dict(parsed)
    for k, v in parsed_flattened.items():
        if isinstance(v, list):
            parsed_flattened[k] = [0 if el is None else el for el in v]

    return parsed_flattened


def get_meta_swainlab(parsed_metadata: dict):
    """
    Convert raw parsing of Swainlab logfile to the metadata interface.

    Input:
    --------
    parsed_metadata: Dict[str, str or int or DataFrame or Dict]
    default['general', 'image_config', 'device_properties', 'group_position', 'group_time', 'group_config']

    Returns:
    --------
    Dictionary with metadata following the standard

    """
    channels = parsed_metadata["image_config"]["Image config"].values.tolist()
    # nframes = int(parsed_metadata["group_time"]["frames"].max())

    # return {"channels": channels, "nframes": nframes}
    return {"channels": channels}


def get_meta_from_legacy(parsed_metadata: dict):
    result = parsed_metadata
    result["channels"] = result["channels/channel"]
    return result


def parse_swainlab_metadata(filedir: t.Union[str, PosixPath]):
    """
    Dispatcher function that determines which parser to use based on the file ending.

    Input:
    --------
    filedir: Directory where the logfile is located.

    Returns:
    --------
    Dictionary with minimal metadata
    """
    filedir = Path(filedir)

    filepath = find_file(filedir, "*.log")
    if filepath:
        raw_parse = parse_from_swainlab_grammar(filepath)
        minimal_meta = get_meta_swainlab(raw_parse)
    else:
        if filedir.is_file() or str(filedir).endswith(".zarr"):
            filedir = filedir.parent
        legacy_parse = parse_logfiles(filedir)
        minimal_meta = (
            get_meta_from_legacy(legacy_parse) if legacy_parse else {}
        )

    return minimal_meta


def dispatch_metadata_parser(filepath: t.Union[str, PosixPath]):
    """
    Function to dispatch different metadata parsers that convert logfiles into a
    basic metadata dictionary. Currently only contains the swainlab log parsers.

    Input:
    --------
    filepath: str existing file containing metadata, or folder containing naming conventions
    """
    parsed_meta = parse_swainlab_metadata(filepath)

    if parsed_meta is None:
        parsed_meta = dir_to_meta

    return parsed_meta


def dir_to_meta(path: PosixPath, suffix="tiff"):
    filenames = list(path.glob(f"*.{suffix}"))

    try:
        # Deduct order from filenames
        dimorder = "".join(
            map(lambda x: x[0], filenames[0].stem.split("_")[1:])
        )

        dim_value = list(
            map(
                lambda f: filename_to_dict_indices(f.stem),
                path.glob("*.tiff"),
            )
        )
        maxes = [max(map(lambda x: x[dim], dim_value)) for dim in dimorder]
        mins = [min(map(lambda x: x[dim], dim_value)) for dim in dimorder]
        _dim_shapes = [
            max_val - min_val + 1 for max_val, min_val in zip(maxes, mins)
        ]

        meta = {
            "size_" + dim: shape for dim, shape in zip(dimorder, _dim_shapes)
        }
    except Exception as e:
        print(
            f"Warning:Metadata: Cannot extract dimensions from filenames. Empty meta set {e}"
        )
        meta = {}

    return meta


def filename_to_dict_indices(stem: str):
    return {
        dim_number[0]: int(dim_number[1:])
        for dim_number in stem.split("_")[1:]
    }



#!/usr/bin/env jupyter
"""
Functions to efficiently merge rows in DataFrames.
"""
import typing as t
from copy import copy

import numpy as np
import pandas as pd
from utils_find_1st import cmp_larger, find_1st

from agora.utils.association import validate_association


def apply_merges(data: pd.DataFrame, merges: np.ndarray):
    """Split data in two, one subset for rows relevant for merging and one
    without them. It uses an array of source tracklets and target tracklets
    to efficiently merge them.

    Parameters
    ----------
    data : pd.DataFrame
        Input DataFrame.
    merges : np.ndarray
        3-D ndarray where dimensions are (X,2,2): nmerges, source-target
        pair and single-cell identifiers, respectively.

    Examples
    --------
    FIXME: Add docs.

    """

    valid_merges, indices = validate_association(
        merges, np.array(list(data.index))
    )

    # Assign non-merged
    merged = data.loc[~indices]

    # Implement the merges and drop source rows.
    if valid_merges.any():
        to_merge = data.loc[indices]
        for target, source in merges[valid_merges]:
            target, source = tuple(target), tuple(source)
            to_merge.loc[target] = join_tracks_pair(
                to_merge.loc[target].values,
                to_merge.loc[source].values,
            )
            to_merge.drop(source, inplace=True)

        merged = pd.concat((merged, to_merge), names=data.index.names)
    return merged


def join_tracks_pair(target: np.ndarray, source: np.ndarray) -> np.ndarray:
    """
    Join two tracks and return the new value of the target.
    """
    target_copy = copy(target)
    end = find_1st(target_copy[::-1], 0, cmp_larger)
    target_copy[-end:] = source[-end:]
    return target_copy
